##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Caching tool implementation.
"""

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from App.Common import rfc1123_date
from App.special_dtml import DTMLFile
from DateTime.DateTime import DateTime
from OFS.Cache import Cache
from OFS.Cache import CacheManager
from OFS.Cache import getVerifiedManagerIds
from OFS.Cache import ZCM_MANAGERS
from OFS.interfaces import IObjectWillBeMovedEvent
from OFS.SimpleItem import SimpleItem
from Persistence import PersistentMapping
from Products.PageTemplates.Expressions import getEngine
from Products.PageTemplates.Expressions import SecureModuleImporter
from zope.component import getUtility
from zope.container.interfaces import IObjectMovedEvent
from zope.interface import implementer

from Products.CMFCore.Expression import Expression
from Products.CMFCore.interfaces import ICachingPolicy
from Products.CMFCore.interfaces import ICachingPolicyManager
from Products.CMFCore.interfaces import IMembershipTool
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.permissions import View
from Products.CMFCore.utils import _dtmldir
from Products.CMFCore.utils import _setCacheHeaders
from Products.CMFCore.utils import _ViewEmulator
from Products.CMFCore.utils import registerToolInterface

# This is lame :(
# This listing is used to decide whether to wrap an object inside a "fake view"
# for the OFS.Cache caching. If it is a view type, no fake view wrap is needed.
VIEW_METATYPES = (
    'Page Template',
    'DTML Method',
    'DTML Document',
    'Filesystem DTML Method',
    'Filesystem Page Template',
    )


def createCPContext(content, view_method, keywords, time=None):
    """
        Construct an expression context for TALES expressions,
        for use by CachingPolicy objects.
    """
    mtool = getUtility(IMembershipTool)
    if mtool.isAnonymousUser():
        member = None
    else:
        member = mtool.getAuthenticatedMember()

    if time is None:
        time = DateTime()

    # The name "content" is deprecated and will go away in CMF 2.0,
    # please use "object" in your policy
    data = {'content': content,
            'object': content,
            'view': view_method,
            'keywords': keywords,
            'request': getattr(content, 'REQUEST', {}),
            'member': member,
            'modules': SecureModuleImporter,
            'nothing': None,
            'time': time}

    return getEngine().getContext(data)


class CPMCache(Cache):
    """ Simple OFS.Cache-implementation
    """
    security = ClassSecurityInfo()

    @security.private
    def ZCache_invalidate(self, ob):
        """ An object is forced out of the cache

        This implementation stores nothing and does not attempt to
        communicate with cache servers, so this is a no-op.
        """
        pass

    @security.private
    def ZCache_get(self, ob, view_name, keywords, mtime_func, default):
        """ An object is retrieved from the cache

        This implementation stores nothing - a no-op.
        """
        pass

    @security.private
    def ZCache_set(self, ob, data, view_name, keywords, mtime_func):
        """ An object is pushed into the cache

        Even though this cache implementation does not cache anything per se,
        this method is used as a suitable hook to activate the real heavy
        lifting done by the CachePolicyManager.
        """
        if ob.meta_type not in VIEW_METATYPES:
            ob = _ViewEmulator().__of__(ob)

        return _setCacheHeaders(ob, extra_context={})

InitializeClass(CPMCache)


@implementer(ICachingPolicy)
class CachingPolicy:
    """
        Represent a single class of cachable objects:

          - class membership is defined by 'predicate', a TALES expression
            with access to the following top-level names:

            'object' -- the object itself

            'view' -- the name of the view method

            'keywords' -- keywords passed to the request

            'request' -- the REQUEST object itself

            'member' -- the authenticated member, or None if anonymous

            'modules' -- usual TALES access-with-import

            'nothing' -- None

            'time' -- A DateTime object for the current date and time

          - mtime_func is used to set the "Last-modified" HTTP response
            header, which is another TALES expression evaluated
            against the same namespace.  If not specified explicitly,
            uses 'object/modified'.  mtime_func is also used in responding
            to conditional GETs.

          - The "Expires" HTTP response header and the "max-age" token of
            the "Cache-control" header will be set using 'max_age_secs',
            if passed;  it should be an integer value in seconds.

          - The "s-maxage" token of the "Cache-control" header will be
            set using 's_max_age_secs', if passed;  it should be an integer
            value in seconds.

          - The "Vary" HTTP response headers will be set if a value is
            provided. The Vary header is described in RFC 2616. In essence,
            it instructs caches that respect this header (such as Squid
            after version 2.4) to distinguish between requests not just by
            the request URL, but also by values found in the headers showing
            in the Vary tag. "Vary: Cookie" would force Squid to also take
            Cookie headers into account when deciding what cached object to
            choose and serve in response to a request.

          - The "ETag" HTTP response header will be set if a value is
            provided. The value is a TALES expression and the result
            after evaluation will be used as the ETag header value.

          - Other tokens will be added to the "Cache-control" HTTP response
            header as follows:

             'no_cache=1' argument => "no-cache" token

             'no_store=1' argument => "no-store" token

             'must_revalidate=1' argument => "must-revalidate" token

             'proxy_revalidate=1' argument => "proxy-revalidate" token

             'public=1' argument => "public" token

             'private=1' argument => "private" token

             'no_transform=1' argument => "no-transform" token

          - The last_modified argument is used to determine whether to add a
            Last-Modified header.  last_modified=1 by default.  There appears
            to be a bug in IE 6 (and possibly other versions) that uses the
            Last-Modified header plus some heuristics rather than the other
            explicit caching headers to determine whether to render content
            from the cache.  If you set, say, max-age=0, must-revalidate and
            have a Last-Modified header some time in the past, IE will
            recognize that the page in cache is stale and will request an
            update from the server BUT if you have a Last-Modified header
            with an older date, will then ignore the update and render from
            the cache, so you may want to disable the Last-Modified header
            when controlling caching using Cache-Control headers.

          - The pre-check and post-check Cache-Control tokens are Microsoft
            proprietary tokens added to IE 5+.  Documentation can be found
            here: http://msdn.microsoft.com/workshop/author/perf/perftips.asp
            Unfortunately these are needed to make IE behave correctly.

    """

    def __init__(self,
                 policy_id,
                 predicate='',
                 mtime_func='',
                 max_age_secs=None,
                 no_cache=0,
                 no_store=0,
                 must_revalidate=0,
                 vary='',
                 etag_func='',
                 s_max_age_secs=None,
                 proxy_revalidate=0,
                 public=0,
                 private=0,
                 no_transform=0,
                 enable_304s=0,
                 last_modified=1,
                 pre_check=None,
                 post_check=None):
        if not predicate:
            predicate = 'python:1'

        if not mtime_func:
            mtime_func = 'object/modified'

        if max_age_secs is not None:
            if str(max_age_secs).strip() == '':
                max_age_secs = None
            else:
                max_age_secs = int(max_age_secs)

        if s_max_age_secs is not None:
            if str(s_max_age_secs).strip() == '':
                s_max_age_secs = None
            else:
                s_max_age_secs = int(s_max_age_secs)

        if pre_check is not None:
            if str(pre_check).strip() == '':
                pre_check = None
            else:
                pre_check = int(pre_check)

        if post_check is not None:
            if str(post_check).strip() == '':
                post_check = None
            else:
                post_check = int(post_check)

        self._policy_id = policy_id
        self._predicate = Expression(text=predicate)
        self._mtime_func = Expression(text=mtime_func)
        self._max_age_secs = max_age_secs
        self._s_max_age_secs = s_max_age_secs
        self._no_cache = int(no_cache)
        self._no_store = int(no_store)
        self._must_revalidate = int(must_revalidate)
        self._proxy_revalidate = int(proxy_revalidate)
        self._public = int(public)
        self._private = int(private)
        self._no_transform = int(no_transform)
        self._vary = vary
        self._etag_func = Expression(text=etag_func)
        self._enable_304s = int(enable_304s)
        self._last_modified = int(last_modified)
        self._pre_check = pre_check
        self._post_check = post_check

    def getPolicyId(self):
        """
        """
        return self._policy_id

    def getPredicate(self):
        """
        """
        return self._predicate.text

    def getMTimeFunc(self):
        """
        """
        return self._mtime_func.text

    def getMaxAgeSecs(self):
        """
        """
        return self._max_age_secs

    def getSMaxAgeSecs(self):
        """
        """
        return getattr(self, '_s_max_age_secs', None)

    def getNoCache(self):
        """
        """
        return self._no_cache

    def getNoStore(self):
        """
        """
        return self._no_store

    def getMustRevalidate(self):
        """
        """
        return self._must_revalidate

    def getProxyRevalidate(self):
        """
        """
        return getattr(self, '_proxy_revalidate', 0)

    def getPublic(self):
        """
        """
        return getattr(self, '_public', 0)

    def getPrivate(self):
        """
        """
        return getattr(self, '_private', 0)

    def getNoTransform(self):
        """
        """
        return getattr(self, '_no_transform', 0)

    def getVary(self):
        """
        """
        return getattr(self, '_vary', '')

    def getETagFunc(self):
        """
        """
        etag_func_text = ''
        etag_func = getattr(self, '_etag_func', None)

        if etag_func is not None:
            etag_func_text = etag_func.text

        return etag_func_text

    def getEnable304s(self):
        """
        """
        return getattr(self, '_enable_304s', 0)

    def getLastModified(self):
        """Should we set the last modified header?"""
        return getattr(self, '_last_modified', 1)

    def getPreCheck(self):
        """
        """
        return getattr(self, '_pre_check', None)

    def getPostCheck(self):
        """
        """
        return getattr(self, '_post_check', None)

    def testPredicate(self, expr_context):
        """ Does this request match our predicate?"""
        return self._predicate(expr_context)

    def getHeaders(self, expr_context):
        """
            Does this request match our predicate?  If so, return a
            sequence of caching headers as ( key, value ) tuples.
            Otherwise, return an empty sequence.
        """
        headers = []

        if self.testPredicate(expr_context):

            if self.getLastModified():
                mtime = self._mtime_func(expr_context)
                if type(mtime) is type(''):
                    mtime = DateTime(mtime)
                if mtime is not None:
                    mtime_str = rfc1123_date(mtime.timeTime())
                    headers.append(('Last-modified', mtime_str))

            control = []

            if self.getMaxAgeSecs() is not None:
                now = expr_context.vars['time']
                exp_time_str = rfc1123_date(now.timeTime()
                                            + self._max_age_secs)
                headers.append(('Expires', exp_time_str))
                control.append('max-age=%d' % self._max_age_secs)

            if self.getSMaxAgeSecs() is not None:
                control.append('s-maxage=%d' % self._s_max_age_secs)

            if self.getNoCache():
                control.append('no-cache')
                # The following is for HTTP 1.0 clients
                headers.append(('Pragma', 'no-cache'))

            if self.getNoStore():
                control.append('no-store')

            if self.getPublic():
                control.append('public')

            if self.getPrivate():
                control.append('private')

            if self.getMustRevalidate():
                control.append('must-revalidate')

            if self.getProxyRevalidate():
                control.append('proxy-revalidate')

            if self.getNoTransform():
                control.append('no-transform')

            pre_check = self.getPreCheck()
            if pre_check is not None:
                control.append('pre-check=%d' % pre_check)

            post_check = self.getPostCheck()
            if post_check is not None:
                control.append('post-check=%d' % post_check)

            if control:
                headers.append(('Cache-control', ', '.join(control)))

            if self.getVary():
                headers.append(('Vary', self._vary))

            if self.getETagFunc():
                headers.append(('ETag', self._etag_func(expr_context)))

        return headers


@implementer(ICachingPolicyManager)
class CachingPolicyManager(SimpleItem, CacheManager):
    """
        Manage the set of CachingPolicy objects for the site;  dispatch
        to them from skin methods.
    """

    id = 'caching_policy_manager'
    meta_type = 'CMF Caching Policy Manager'
    _isCacheManager = 1 # Dead chicken. Yum.

    security = ClassSecurityInfo()

    def __init__(self):
        self._policy_ids = ()
        self._policies = PersistentMapping()

    #
    #   ZMI
    #
    manage_options = (
        ({'label': 'Policies', 'action': 'manage_cachingPolicies',
          'help': ('CMFCore', 'CPMPolicies.stx')},) +
        CacheManager.manage_options +
        SimpleItem.manage_options)

    security.declareProtected(ManagePortal, 'manage_cachingPolicies')
    manage_cachingPolicies = DTMLFile('cachingPolicies', _dtmldir)

    @security.public
    def listPolicies(self):
        """List '(id, (policy, typeObjectName))' tuples for all policies.
        """
        return tuple([ (id, self._policies[id]) for id in self._policy_ids ])

    @security.protected(ManagePortal)
    def addPolicy(self,
                  policy_id,
                  predicate, # TALES expr (def. 'python:1')
                  mtime_func, # TALES expr (def. 'object/modified')
                  max_age_secs, # integer, seconds (def. 0)
                  no_cache, # boolean (def. 0)
                  no_store, # boolean (def. 0)
                  must_revalidate, # boolean (def. 0)
                  vary, # string value
                  etag_func, # TALES expr (def. '')
                  REQUEST=None,
                  s_max_age_secs=None, # integer, seconds (def. None)
                  proxy_revalidate=0, # boolean (def. 0)
                  public=0, # boolean (def. 0)
                  private=0, # boolean (def. 0)
                  no_transform=0, # boolean (def. 0)
                  enable_304s=0, # boolean (def. 0)
                  last_modified=1, # boolean (def. 1)
                  pre_check=None, # integer, default None
                  post_check=None): # integer, default None
        """
            Add a caching policy.
        """
        if max_age_secs is None or str(max_age_secs).strip() == '':
            max_age_secs = None
        else:
            max_age_secs = int(max_age_secs)

        if s_max_age_secs is None or str(s_max_age_secs).strip() == '':
            s_max_age_secs = None
        else:
            s_max_age_secs = int(s_max_age_secs)

        if pre_check is None or str(pre_check).strip() == '':
            pre_check = None
        else:
            pre_check = int(pre_check)

        if post_check is None or str(post_check).strip() == '':
            post_check = None
        else:
            post_check = int(post_check)

        self._addPolicy(policy_id,
                        predicate,
                        mtime_func,
                        max_age_secs,
                        no_cache,
                        no_store,
                        must_revalidate,
                        vary,
                        etag_func,
                        s_max_age_secs,
                        proxy_revalidate,
                        public,
                        private,
                        no_transform,
                        enable_304s,
                        last_modified,
                        pre_check,
                        post_check)
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url()
                                         + '/manage_cachingPolicies'
                                         + '?manage_tabs_message='
                                         + 'Policy+added.')

    @security.protected(ManagePortal)
    def updatePolicy(self,
                     policy_id,
                     predicate, # TALES expr (def. 'python:1')
                     mtime_func, # TALES expr (def. 'object/modified')
                     max_age_secs, # integer, seconds (def. 0)
                     no_cache, # boolean (def. 0)
                     no_store, # boolean (def. 0)
                     must_revalidate, # boolean (def. 0)
                     vary, # string value
                     etag_func, # TALES expr (def. '')
                     REQUEST=None,
                     s_max_age_secs=None, # integer, seconds (def. 0)
                     proxy_revalidate=0, # boolean (def. 0)
                     public=0, # boolean (def. 0)
                     private=0, # boolean (def. 0)
                     no_transform=0, # boolean (def. 0)
                     enable_304s=0, # boolean (def. 0)
                     last_modified=1, # boolean (def. 1)
                     pre_check=0, # integer, default=None
                     post_check=0): # integer, default=None
        """
            Update a caching policy.
        """
        if max_age_secs is None or str(max_age_secs).strip() == '':
            max_age_secs = None
        else:
            max_age_secs = int(max_age_secs)

        if s_max_age_secs is None or str(s_max_age_secs).strip() == '':
            s_max_age_secs = None
        else:
            s_max_age_secs = int(s_max_age_secs)

        if pre_check is None or str(pre_check).strip() == '':
            pre_check = None
        else:
            pre_check = int(pre_check)

        if post_check is None or str(post_check).strip() == '':
            post_check = None
        else:
            post_check = int(post_check)

        self._updatePolicy(policy_id,
                           predicate,
                           mtime_func,
                           max_age_secs,
                           no_cache,
                           no_store,
                           must_revalidate,
                           vary,
                           etag_func,
                           s_max_age_secs,
                           proxy_revalidate,
                           public,
                           private,
                           no_transform,
                           enable_304s,
                           last_modified,
                           pre_check,
                           post_check)
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url()
                                         + '/manage_cachingPolicies'
                                         + '?manage_tabs_message='
                                         + 'Policy+updated.')

    @security.protected(ManagePortal)
    def movePolicyUp(self, policy_id, REQUEST=None):
        """
            Move a caching policy up in the list.
        """
        policy_ids = list(self._policy_ids)
        ndx = policy_ids.index(policy_id)
        if ndx == 0:
            msg = "Policy+already+first."
        else:
            self._reorderPolicy(policy_id, ndx - 1)
            msg = "Policy+moved."
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url()
                                         + '/manage_cachingPolicies'
                                         + '?manage_tabs_message=%s' % msg)

    @security.protected(ManagePortal)
    def movePolicyDown(self, policy_id, REQUEST=None):
        """
            Move a caching policy down in the list.
        """
        policy_ids = list(self._policy_ids)
        ndx = policy_ids.index(policy_id)
        if ndx == len(policy_ids) - 1:
            msg = "Policy+already+last."
        else:
            self._reorderPolicy(policy_id, ndx + 1)
            msg = "Policy+moved."
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url()
                                         + '/manage_cachingPolicies'
                                         + '?manage_tabs_message=%s' % msg)

    @security.protected(ManagePortal)
    def removePolicy(self, policy_id, REQUEST=None):
        """
            Remove a caching policy.
        """
        self._removePolicy(policy_id)
        if REQUEST is not None:
            REQUEST['RESPONSE'].redirect(self.absolute_url()
                                      + '/manage_cachingPolicies'
                                      + '?manage_tabs_message=Policy+removed.')

    #
    #   Policy manipulation methods.
    #
    @security.private
    def _addPolicy(self,
                   policy_id,
                   predicate,
                   mtime_func,
                   max_age_secs,
                   no_cache,
                   no_store,
                   must_revalidate,
                   vary,
                   etag_func,
                   s_max_age_secs=None,
                   proxy_revalidate=0,
                   public=0,
                   private=0,
                   no_transform=0,
                   enable_304s=0,
                   last_modified=1,
                   pre_check=None,
                   post_check=None):
        """
            Add a policy to our registry.
        """
        policy_id = str(policy_id).strip()

        if not policy_id:
            raise ValueError("Policy ID is required!")

        if policy_id in self._policy_ids:
            raise KeyError("Policy %s already exists!" % policy_id)

        self._policies[policy_id] = CachingPolicy(policy_id,
                                                  predicate,
                                                  mtime_func,
                                                  max_age_secs,
                                                  no_cache,
                                                  no_store,
                                                  must_revalidate,
                                                  vary,
                                                  etag_func,
                                                  s_max_age_secs,
                                                  proxy_revalidate,
                                                  public,
                                                  private,
                                                  no_transform,
                                                  enable_304s,
                                                  last_modified,
                                                  pre_check,
                                                  post_check)
        idlist = list(self._policy_ids)
        idlist.append(policy_id)
        self._policy_ids = tuple(idlist)

    @security.private
    def _updatePolicy(self,
                      policy_id,
                      predicate,
                      mtime_func,
                      max_age_secs,
                      no_cache,
                      no_store,
                      must_revalidate,
                      vary,
                      etag_func,
                      s_max_age_secs=None,
                      proxy_revalidate=0,
                      public=0,
                      private=0,
                      no_transform=0,
                      enable_304s=0,
                      last_modified=1,
                      pre_check=None,
                      post_check=None):
        """
            Update a policy in our registry.
        """
        if policy_id not in self._policy_ids:
            raise KeyError("Policy %s does not exist!" % policy_id)

        self._policies[policy_id] = CachingPolicy(policy_id,
                                                  predicate,
                                                  mtime_func,
                                                  max_age_secs,
                                                  no_cache,
                                                  no_store,
                                                  must_revalidate,
                                                  vary,
                                                  etag_func,
                                                  s_max_age_secs,
                                                  proxy_revalidate,
                                                  public,
                                                  private,
                                                  no_transform,
                                                  enable_304s,
                                                  last_modified,
                                                  pre_check,
                                                  post_check)

    @security.private
    def _reorderPolicy(self, policy_id, newIndex):
        """
            Reorder a policy in our registry.
        """
        if policy_id not in self._policy_ids:
            raise KeyError("Policy %s does not exist!" % policy_id)

        idlist = list(self._policy_ids)
        ndx = idlist.index(policy_id)
        pred = idlist[ndx]
        idlist = idlist[:ndx] + idlist[ndx + 1:]
        idlist.insert(newIndex, pred)
        self._policy_ids = tuple(idlist)

    @security.private
    def _removePolicy(self, policy_id):
        """
            Remove a policy from our registry.
        """
        if policy_id not in self._policy_ids:
            raise KeyError("Policy %s does not exist!" % policy_id)

        del self._policies[policy_id]
        idlist = list(self._policy_ids)
        ndx = idlist.index(policy_id)
        idlist = idlist[:ndx] + idlist[ndx + 1:]
        self._policy_ids = tuple(idlist)

    #
    #   'portal_caching' interface methods
    #
    @security.protected(View)
    def getHTTPCachingHeaders(self, content, view_method, keywords, time=None):
        """
            Return a list of HTTP caching headers based on 'content',
            'view_method', and 'keywords'.
        """
        context = createCPContext(content, view_method, keywords, time=time)
        for _policy_id, policy in self.listPolicies():

            headers = policy.getHeaders(context)
            if headers:
                return headers

        return ()

    @security.protected(View)
    def getModTimeAndETag(self, content, view_method, keywords, time=None):
        """ Return the modification time and ETag for the content object,
            view method, and keywords as the tuple (modification_time, etag,
            set_last_modified_header), where modification_time is a DateTime,
            or None.
        """
        context = createCPContext(content, view_method, keywords, time=time)
        for _policy_id, policy in self.listPolicies():
            if policy.getEnable304s() and policy.testPredicate(context):

                last_modified = policy._mtime_func(context)
                if type(last_modified) is type(''):
                    last_modified = DateTime(last_modified)

                content_etag = None
                if policy.getETagFunc():
                    content_etag = policy._etag_func(context)

                return (last_modified, content_etag, policy.getLastModified())

        return None

    #
    # OFS.CacheManager API
    #
    @security.private
    def ZCacheManager_getCache(self):
        """ Retrieve a cache object
        """
        cache = getattr(self, '_cache', None)

        if cache is None:
            self._cache = CPMCache()
            cache = self._cache

        return cache

InitializeClass(CachingPolicyManager)
registerToolInterface('caching_policy_manager', ICachingPolicyManager)


def handleCachingPolicyManagerEvent(ob, event):
    """ Event subscriber for (un)registering a CPM as CacheManager
    """
    if not ICachingPolicyManager.providedBy(ob):
        return

    if IObjectMovedEvent.providedBy(event):
        if event.newParent is not None:
            ids = getVerifiedManagerIds(event.newParent)
            id = ob.getId()
            if id not in ids:
                setattr(event.newParent, ZCM_MANAGERS, ids + (id,))

    elif IObjectWillBeMovedEvent.providedBy(event):
        if event.oldParent is not None:
            ids = list(getVerifiedManagerIds(event.oldParent))
            id = ob.getId()
            if id in ids:
                ids.remove(id)
                setattr(event.oldParent, ZCM_MANAGERS, tuple(ids))


def manage_addCachingPolicyManager(self, REQUEST=None):
    """
        Add a CPM to self.
    """
    id = CachingPolicyManager.id
    mgr = CachingPolicyManager()
    self._setObject(id, mgr)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(self.absolute_url()
                        + '/manage_main'
                        + '?manage_tabs_message=Caching+Policy+Manager+added.')



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Actions tool node adapter unit tests.
"""

import unittest

from Acquisition import aq_parent
from Acquisition import Implicit
from OFS.OrderedFolder import OrderedFolder
from zope.component import getSiteManager
from zope.interface import implementer

from Products.CMFCore.ActionProviderBase import ActionProviderBase
from Products.CMFCore.interfaces import IActionProvider
from Products.CMFCore.interfaces import IActionsTool
from Products.CMFCore.interfaces import IMembershipTool
from Products.CMFCore.testing import ExportImportZCMLLayer
from Products.CMFCore.tests.base.dummy import DummySite
from Products.GenericSetup.testing import BodyAdapterTestCase
from Products.GenericSetup.testing import NodeAdapterTestCase
from Products.GenericSetup.tests.common import BaseRegistryTests
from Products.GenericSetup.tests.common import DummyExportContext
from Products.GenericSetup.tests.common import DummyImportContext

_ACTION_XML = b"""\
<object name="foo_action" meta_type="CMF Action">
 <property name="title">Foo</property>
 <property name="description"></property>
 <property name="url_expr">string:${object_url}/foo</property>
 <property name="link_target"></property>
 <property name="icon_expr"></property>
 <property name="available_expr">python:1</property>
 <property name="permissions"/>
 <property name="visible">True</property>
</object>
"""

_ACTIONCATEGORY_XML = b"""\
<object name="foo_category" meta_type="CMF Action Category">
 <property name="title"></property>
 <object name="foo_action" meta_type="CMF Action">
  <property name="title"></property>
  <property name="description"></property>
  <property name="url_expr"></property>
  <property name="link_target"></property>
  <property name="icon_expr"></property>
  <property name="available_expr"></property>
  <property name="permissions"/>
  <property name="visible">True</property>
 </object>
</object>
"""

_ACTIONSTOOL_BODY = b"""\
<?xml version="1.0" encoding="utf-8"?>
<object name="portal_actions" meta_type="CMF Actions Tool"
   xmlns:i18n="http://xml.zope.org/namespaces/i18n">
 <action-provider name="portal_actions"/>
 <object name="foo_category" meta_type="CMF Action Category">
  <property name="title"></property>
  <object name="foo_action" meta_type="CMF Action" i18n:domain="foo_domain">
   <property name="title" i18n:translate=""></property>
   <property name="description" i18n:translate=""></property>
   <property name="url_expr"></property>
   <property name="link_target"></property>
   <property name="icon_expr"></property>
   <property name="available_expr"></property>
   <property name="permissions"/>
   <property name="visible">True</property>
  </object>
 </object>
</object>
"""

_EMPTY_EXPORT = b"""\
<?xml version="1.0"?>
<object name="portal_actions" meta_type="CMF Actions Tool"
   xmlns:i18n="http://xml.zope.org/namespaces/i18n">
 <action-provider name="portal_actions"/>
</object>
"""

_OLD_EXPORT = b"""\
<?xml version="1.0"?>
<object name="portal_actions" meta_type="CMF Actions Tool"
   xmlns:i18n="http://xml.zope.org/namespaces/i18n">
 <action-provider name="portal_actions">
  <action action_id="baz"
          title="Baz"
          url_expr="string:${object_url}/baz"
          condition_expr="python:1"
          category="dummy"
          visible="True"/>
 </action-provider>
 <action-provider name="portal_foo">
  <action action_id="foo"
          title="Foo"
          url_expr="string:${object_url}/foo"
          condition_expr="python:1"
          category="dummy"
          visible="True"/>
 </action-provider>
 <action-provider name="portal_bar">
  <action action_id="bar"
          title="Bar"
          url_expr="string:${object_url}/bar"
          condition_expr="python:0"
          category="dummy"
          visible="False">
   <permission>Manage portal</permission>
  </action>
 </action-provider>
</object>
"""

_NORMAL_EXPORT = """\
<?xml version="1.0" encoding="utf-8"?>
<object name="portal_actions" meta_type="CMF Actions Tool"
   xmlns:i18n="http://xml.zope.org/namespaces/i18n">
 <action-provider name="portal_actions">
  <action action_id="baz"
          title="Baz"
          url_expr="string:${object_url}/baz"
          condition_expr="python:1"
          category="dummy"
          visible="True"/>
 </action-provider>
 <action-provider name="portal_foo"/>
 <action-provider name="portal_bar"/>
</object>
"""

_NEWSYTLE_EXPORT = b"""\
<?xml version="1.0"?>
<object name="portal_actions" meta_type="CMF Actions Tool"
   xmlns:i18n="http://xml.zope.org/namespaces/i18n">
 <action-provider name="portal_actions"/>
 <action-provider name="portal_foo"/>
 <action-provider name="portal_bar"/>
 <object name="dummy" meta_type="CMF Action Category">
  <property name="title"></property>
  <object name="baz" meta_type="CMF Action">
   <property name="title">Baz</property>
   <property name="description"></property>
   <property name="url_expr">string:${object_url}/baz</property>
   <property name="link_target"></property>
   <property name="icon_expr"></property>
   <property name="available_expr">python:1</property>
   <property name="permissions"></property>
   <property name="visible">True</property>
  </object>
 </object>
</object>
"""

_I18N_IMPORT = """\
<?xml version="1.0"?>
<object name="portal_actions" meta_type="CMF Actions Tool"
   xmlns:i18n="http://xml.zope.org/namespaces/i18n">
 <action-provider name="portal_actions"/>
 <object name="dummy" meta_type="CMF Action Category">
  <property name="title"></property>
  <object name="foo" meta_type="CMF Action" i18n:domain="foo_domain">
   <property name="title" i18n:translate="">Foo</property>
   <property name="description" i18n:translate=""></property>
   <property name="url_expr">string:${object_url}/foo</property>
   <property name="link_target"></property>
   <property name="icon_expr"></property>
   <property name="available_expr">python:1</property>
   <property name="permissions"></property>
   <property name="visible">True</property>
  </object>
 </object>
</object>
"""

_INSERT_IMPORT = """\
<?xml version="1.0"?>
<object name="portal_actions">
 <object name="dummy">
 <object name="spam" meta_type="CMF Action" insert-before="*">
  <property name="title">Spam</property>
  <property name="description"></property>
  <property name="url_expr">string:${object_url}/spam</property>
  <property name="icon_expr">string:spam_icon.png</property>
  <property name="available_expr"></property>
  <property name="permissions">
   <element value="View" /></property>
  <property name="visible">True</property>
 </object>
 <object name="baz" insert-after="*">
  <property name="icon_expr">string:baz_icon.png</property>
 </object>
 </object>
</object>
"""

_REMOVE_IMPORT = """\
<?xml version="1.0"?>
<object name="portal_actions">
 <action-provider name="portal_actions" remove=""/>
 <action-provider name="not_existing" remove=""/>
 <action-provider name="portal_bar" remove=""/>
</object>
"""


@implementer(IActionProvider)
class DummyTool(OrderedFolder, ActionProviderBase):
    pass


class DummyUser(Implicit):

    def getId(self):
        return 'dummy'


class DummyMembershipTool(DummyTool):

    def isAnonymousUser(self):
        return False

    def getAuthenticatedMember(self):
        return DummyUser().__of__(aq_parent(self))


@implementer(IActionsTool)
class DummyActionsTool(DummyTool):

    id = 'portal_actions'
    meta_type = 'CMF Actions Tool'

    def __init__(self):
        self._providers = []

    def addActionProvider(self, provider_name):
        self._providers.append(provider_name)

    def listActionProviders(self):
        return self._providers[:]

    def deleteActionProvider(self, provider_name):
        self._providers = [ x for x in self._providers if x != provider_name ]


class ActionNodeAdapterTests(NodeAdapterTestCase, unittest.TestCase):

    layer = ExportImportZCMLLayer

    def _getTargetClass(self):
        from Products.CMFCore.exportimport.actions import ActionNodeAdapter

        return ActionNodeAdapter

    def _populate(self, obj):
        obj._setPropValue('title', 'Foo')
        obj._setPropValue('url_expr', 'string:${object_url}/foo')
        obj._setPropValue('available_expr', 'python:1')

    def _verifyImport(self, obj):
        self.assertEqual(type(obj.title), str)
        self.assertEqual(obj.title, 'Foo')
        self.assertEqual(type(obj.description), str)
        self.assertEqual(obj.description, '')
        self.assertEqual(type(obj.url_expr), str)
        self.assertEqual(obj.url_expr, 'string:${object_url}/foo')
        self.assertEqual(type(obj.icon_expr), str)
        self.assertEqual(obj.icon_expr, '')
        self.assertEqual(type(obj.available_expr), str)
        self.assertEqual(obj.available_expr, 'python:1')
        self.assertEqual(type(obj.permissions), tuple)
        self.assertEqual(obj.permissions, ())
        self.assertEqual(type(obj.visible), bool)
        self.assertEqual(obj.visible, True)

    def setUp(self):
        from Products.CMFCore.ActionInformation import Action

        self._obj = Action('foo_action')
        self._XML = _ACTION_XML


class ActionCategoryNodeAdapterTests(NodeAdapterTestCase, unittest.TestCase):

    layer = ExportImportZCMLLayer

    def _getTargetClass(self):
        from Products.CMFCore.exportimport.actions \
                import ActionCategoryNodeAdapter

        return ActionCategoryNodeAdapter

    def _populate(self, obj):
        from Products.CMFCore.ActionInformation import Action

        obj._setObject('foo_action', Action('foo_action'))

    def _verifyImport(self, obj):
        self.assertEqual(type(obj.title), str)
        self.assertEqual(obj.title, '')

    def setUp(self):
        from Products.CMFCore.ActionInformation import ActionCategory

        self._obj = ActionCategory('foo_category')
        self._XML = _ACTIONCATEGORY_XML


class ActionsToolXMLAdapterTests(BodyAdapterTestCase, unittest.TestCase):

    layer = ExportImportZCMLLayer

    def _getTargetClass(self):
        from Products.CMFCore.exportimport.actions \
                import ActionsToolXMLAdapter

        return ActionsToolXMLAdapter

    def _populate(self, obj):
        from Products.CMFCore.ActionInformation import Action
        from Products.CMFCore.ActionInformation import ActionCategory

        obj._setObject('foo_category', ActionCategory('foo_category'))
        obj.action_providers = ('portal_actions',)
        obj.foo_category._setObject('foo_action', Action('foo_action'))
        obj.foo_category.foo_action.i18n_domain = 'foo_domain'

    def _verifyImport(self, obj):
        self.assertEqual(type(obj.action_providers), tuple)
        self.assertEqual(obj.action_providers, ('portal_actions',))
        self.assertEqual(type(obj.action_providers[0]), str)
        self.assertEqual(obj.action_providers[0], 'portal_actions')

    def setUp(self):
        from Products.CMFCore.ActionsTool import ActionsTool

        self._obj = ActionsTool('portal_actions')
        self._BODY = _ACTIONSTOOL_BODY
        getSiteManager().registerUtility(self._obj, IActionsTool)


class _ActionSetup(BaseRegistryTests):

    def _initSite(self, foo=2, bar=2):
        site = DummySite('site')
        sm = getSiteManager()
        sm.registerUtility(DummyMembershipTool(), IMembershipTool)
        atool = DummyActionsTool()
        atool.addActionProvider('portal_actions')
        sm.registerUtility(atool, IActionsTool)

        if foo > 0:
            site.portal_foo = DummyTool()

        if foo > 1:
            site.portal_foo.addAction(id='foo',
                                      name='Foo',
                                      action='foo',
                                      condition='python:1',
                                      permission=(),
                                      category='dummy',
                                      visible=1)
            atool.addActionProvider('portal_foo')

        if bar > 0:
            site.portal_bar = DummyTool()

        if bar > 1:
            site.portal_bar.addAction(id='bar',
                                      name='Bar',
                                      action='bar',
                                      condition='python:0',
                                      permission=('Manage portal',),
                                      category='dummy',
                                      visible=0)
            atool.addActionProvider('portal_bar')

        return site, atool


class exportActionProvidersTests(_ActionSetup):

    layer = ExportImportZCMLLayer

    def test_unchanged(self):
        from Products.CMFCore.exportimport.actions \
                import exportActionProviders

        site, _atool = self._initSite(0, 0)
        context = DummyExportContext(site)
        exportActionProviders(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'actions.xml')
        self._compareDOM(text.decode('utf8'), _EMPTY_EXPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_normal(self):
        from Products.CMFCore.exportimport.actions \
                import exportActionProviders

        site, atool = self._initSite()
        # Set up an old action for added difficulty
        atool.addAction(id='baz',
                        name='Baz',
                        action='baz',
                        condition='python:1',
                        permission=(),
                        category='dummy',
                        visible=1)

        context = DummyExportContext(site)
        exportActionProviders(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'actions.xml')
        self._compareDOM(text.decode('utf-8'), _NORMAL_EXPORT)
        self.assertEqual(content_type, 'text/xml')


class importActionProvidersTests(_ActionSetup):

    layer = ExportImportZCMLLayer

    def test_empty_default_purge(self):
        from Products.CMFCore.exportimport.actions \
                import importActionProviders

        site, atool = self._initSite(2, 0)

        self.assertEqual(len(atool.listActionProviders()), 2)
        self.assertTrue('portal_foo' in atool.listActionProviders())
        self.assertTrue('portal_actions' in atool.listActionProviders())

        context = DummyImportContext(site)
        context._files['actions.xml'] = _EMPTY_EXPORT
        importActionProviders(context)

        self.assertEqual(len(atool.listActionProviders()), 1)
        self.assertFalse('portal_foo' in atool.listActionProviders())
        self.assertTrue('portal_actions' in atool.listActionProviders())
        self.assertEqual(len(atool.objectIds()), 0)

    def test_empty_explicit_purge(self):
        from Products.CMFCore.exportimport.actions \
                import importActionProviders

        site, atool = self._initSite(2, 0)

        self.assertEqual(len(atool.listActionProviders()), 2)
        self.assertTrue('portal_foo' in atool.listActionProviders())
        self.assertTrue('portal_actions' in atool.listActionProviders())

        context = DummyImportContext(site, True)
        context._files['actions.xml'] = _EMPTY_EXPORT
        importActionProviders(context)

        self.assertEqual(len(atool.listActionProviders()), 1)
        self.assertFalse('portal_foo' in atool.listActionProviders())
        self.assertTrue('portal_actions' in atool.listActionProviders())
        self.assertEqual(len(atool.objectIds()), 0)

    def test_empty_skip_purge(self):
        from Products.CMFCore.exportimport.actions \
                import importActionProviders

        site, atool = self._initSite(2, 0)

        self.assertEqual(len(atool.listActionProviders()), 2)
        self.assertTrue('portal_foo' in atool.listActionProviders())
        self.assertTrue('portal_actions' in atool.listActionProviders())

        context = DummyImportContext(site, False)
        context._files['actions.xml'] = _EMPTY_EXPORT
        importActionProviders(context)

        self.assertEqual(len(atool.listActionProviders()), 2)
        self.assertTrue('portal_foo' in atool.listActionProviders())
        self.assertTrue('portal_actions' in atool.listActionProviders())

    def test_normal(self):
        from Products.CMFCore.exportimport.actions \
                import exportActionProviders
        from Products.CMFCore.exportimport.actions \
                import importActionProviders

        site, atool = self._initSite(1, 1)
        foo = site.portal_foo
        bar = site.portal_bar

        self.assertEqual(len(atool.listActionProviders()), 1)
        self.assertFalse('portal_foo' in atool.listActionProviders())
        self.assertFalse(foo.listActions())
        self.assertFalse('portal_bar' in atool.listActionProviders())
        self.assertFalse(bar.listActions())
        self.assertTrue('portal_actions' in atool.listActionProviders())

        context = DummyImportContext(site)
        context._files['actions.xml'] = _OLD_EXPORT
        importActionProviders(context)

        self.assertEqual(len(atool.listActionProviders()), 3)
        self.assertTrue('portal_bar' in atool.listActionProviders())
        self.assertTrue('portal_foo' in atool.listActionProviders())
        self.assertTrue('portal_actions' in atool.listActionProviders())

        self.assertEqual(len(atool.objectIds()), 1)
        self.assertTrue('dummy' in atool.objectIds())
        # Only one action appears. The importer only deals with actions
        # defined by the actions tool. Other tools are responsible for
        # exporting/importing actions themselves.
        self.assertEqual(len(atool.dummy.objectIds()), 1)
        self.assertTrue('baz' in atool.dummy.objectIds())
        self.assertFalse(foo.listActions())
        self.assertFalse(bar.listActions())

        # complete the roundtrip
        context = DummyExportContext(site)
        exportActionProviders(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'actions.xml')
        self._compareDOM(text.decode('utf8'), _NEWSYTLE_EXPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_i18n(self):
        from Products.CMFCore.exportimport.actions \
                import exportActionProviders
        from Products.CMFCore.exportimport.actions \
                import importActionProviders

        site, atool = self._initSite(0, 0)

        context = DummyImportContext(site)
        context._files['actions.xml'] = _I18N_IMPORT
        importActionProviders(context)

        self.assertEqual(len(atool.listActionProviders()), 1)
        self.assertEqual(atool.objectIds(), ['dummy'])
        self.assertEqual(atool.dummy.objectIds(), ['foo'])
        self.assertEqual(atool.dummy.foo.i18n_domain, 'foo_domain')

        # complete the roundtrip
        context = DummyExportContext(site)
        exportActionProviders(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'actions.xml')
        self._compareDOM(text.decode('utf8'), _I18N_IMPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_insert_skip_purge(self):
        from Products.CMFCore.exportimport.actions \
                import importActionProviders

        site, atool = self._initSite(0, 0)

        context = DummyImportContext(site)
        context._files['actions.xml'] = _NEWSYTLE_EXPORT
        importActionProviders(context)

        self.assertEqual(len(atool.listActionProviders()), 3)
        self.assertEqual(atool.objectIds(), ['dummy'])
        self.assertEqual(atool.dummy.objectIds(), ['baz'])
        self.assertEqual(atool.dummy.baz.icon_expr, '')

        context = DummyImportContext(site, False)
        context._files['actions.xml'] = _INSERT_IMPORT
        importActionProviders(context)

        self.assertEqual(len(atool.listActionProviders()), 3)
        self.assertEqual(atool.objectIds(), ['dummy'])
        self.assertEqual(atool.dummy.objectIds(), ['spam', 'baz'])
        self.assertEqual(atool.dummy.baz.icon_expr, 'string:baz_icon.png')

    def test_remove_skip_purge(self):
        from Products.CMFCore.exportimport.actions \
                import importActionProviders

        site, atool = self._initSite(2, 2)

        self.assertEqual(atool.listActionProviders(),
                         ['portal_actions', 'portal_foo', 'portal_bar'])

        context = DummyImportContext(site, False)
        context._files['actions.xml'] = _REMOVE_IMPORT
        importActionProviders(context)

        self.assertEqual(atool.listActionProviders(), ['portal_foo'])


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(ActionNodeAdapterTests),
        unittest.makeSuite(ActionCategoryNodeAdapterTests),
        unittest.makeSuite(ActionsToolXMLAdapterTests),
        unittest.makeSuite(exportActionProvidersTests),
        unittest.makeSuite(importActionProvidersTests),
        ))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for ActionInformation module.
"""

import unittest
import Testing

from OFS.Folder import manage_addFolder
from Products.PythonScripts.PythonScript import manage_addPythonScript
from zope.component import getSiteManager
from zope.interface.verify import verifyClass
from zope.testing.cleanup import cleanUp

from Products.CMFCore.Expression import createExprContext
from Products.CMFCore.Expression import Expression
from Products.CMFCore.interfaces import IMembershipTool
from Products.CMFCore.testing import FunctionalZCMLLayer
from Products.CMFCore.tests.base.dummy import DummyContent
from Products.CMFCore.tests.base.dummy import DummySite
from Products.CMFCore.tests.base.dummy import DummyTool as DummyMembershipTool
from Products.CMFCore.tests.base.testcase import SecurityTest
from Products.CMFCore.tests.base.testcase import TransactionalTest


class ActionCategoryTests(unittest.TestCase):

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ActionInformation import ActionCategory

        return ActionCategory(*args, **kw)

    def test_interfaces(self):
        from Products.CMFCore.ActionInformation import ActionCategory
        from Products.CMFCore.interfaces import IActionCategory

        verifyClass(IActionCategory, ActionCategory)

    def test_listActions(self):
        from Products.CMFCore.ActionInformation import Action

        ac = self._makeOne('foo')
        self.assertEqual(ac.listActions(), ())

        baz = Action('baz')
        ac._setObject('baz', baz)
        self.assertEqual(ac.listActions(), (baz,))


class ActionTests(unittest.TestCase):

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ActionInformation import Action

        return Action(*args, **kw)

    def test_interfaces(self):
        from Products.CMFCore.ActionInformation import Action
        from Products.CMFCore.interfaces import IAction

        verifyClass(IAction, Action)

    def test_getInfoData_empty(self):
        WANTED = ({'available': True, 'category': '', 'description': '',
                   'id': 'foo', 'icon': '', 'permissions': (), 'title': '',
                   'url': '', 'visible': True, 'link_target': None}, [])
        a = self._makeOne('foo')
        self.assertEqual(a.getInfoData(), WANTED)

    def test_getInfoData_normal(self):
        a = self._makeOne('foo',
                          title='Foo Title',
                          description='Foo description.',
                          url_expr='string:${object_url}/foo_url',
                          icon_expr='string:foo_icon',
                          available_expr='',
                          permissions=('View',),
                          visible=False,
                          link_target='_top')
        WANTED = ({'available': True, 'category': '',
                   'description': 'Foo description.',
                   'id': 'foo', 'icon': a.icon_expr_object,
                   'permissions': ('View',), 'title': 'Foo Title',
                   'url': a.url_expr_object, 'visible': False,
                   'link_target': a.link_target},
                  ['url', 'icon'])
        self.assertEqual(a.getInfoData(), WANTED)

    def test_manage_propertiesForm_allows_adding(self):
        from OFS.SimpleItem import SimpleItem

        def _header(*args, **kw):
            return 'HEADER'

        def _footer(*args, **kw):
            return 'HEADER'

        container = SimpleItem()

        container.REQUEST = request = DummyRequest()
        request.set('manage_page_header', _header)
        request.set('manage_page_footer', _footer)
        request.set('BASEPATH1', '/one/two')
        setattr(request, 'URL1', '/one/two')
        request._steps = ['one', 'two']

        prd = {'ac_permissions': ('a', 'b')}
        container._getProductRegistryData = prd.get

        a = self._makeOne('extensible').__of__(container)
        form_html = a.manage_propertiesForm(request)

        self.assertTrue('value=" Add "' in form_html)

    def test_clearExprObjects(self):
        """When a *_expr property is set, a *_expr_object attribute is
        also set which should also be cleared when the *_expr is
        cleared."""
        a = self._makeOne('foo',
                          title='Foo Title',
                          description='Foo description.',
                          url_expr='string:${object_url}/foo_url',
                          icon_expr='string:foo_icon',
                          available_expr='',
                          permissions=('View',),
                          visible=False,
                          link_target='_top')
        self.assertTrue(hasattr(a, 'icon_expr_object'))
        self.assertTrue(hasattr(a, 'url_expr_object'))
        self.assertFalse(hasattr(a, 'available_expr_object'))
        a.manage_changeProperties(
            icon_expr='', url_expr='', available_expr='')
        self.assertFalse(hasattr(a, 'icon_expr_object'))
        self.assertFalse(hasattr(a, 'url_expr_object'))
        self.assertFalse(hasattr(a, 'available_expr_object'))


class DummyRequest:

    def __init__(self):
        self._data = {}

    def set(self, k, v):
        self._data[k] = v

    def get(self, k, default):
        return self._data.get(k, default)

    def __getitem__(self, k):
        return self._data[k]

    def __len__(self):
        return len(self._data)


class ActionInfoTests(unittest.TestCase):

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ActionInformation import ActionInfo

        return ActionInfo(*args, **kw)

    def test_interfaces(self):
        from Products.CMFCore.ActionInformation import ActionInfo
        from Products.CMFCore.interfaces import IActionInfo

        verifyClass(IActionInfo, ActionInfo)

    def test_create_from_Action(self):
        from Products.CMFCore.ActionInformation import Action

        WANTED = {'allowed': True, 'available': True, 'category': '',
                  'description': '', 'icon': '', 'id': 'foo', 'title': '',
                  'url': '', 'visible': True, 'link_target': None}

        action = Action(id='foo')
        ec = None
        ai = self._makeOne(action, ec)

        self.assertEqual(ai['id'], WANTED['id'])
        self.assertEqual(ai['title'], WANTED['title'])
        self.assertEqual(ai['description'], WANTED['description'])
        self.assertEqual(ai['url'], WANTED['url'])
        self.assertEqual(ai['category'], WANTED['category'])
        self.assertEqual(ai['visible'], WANTED['visible'])
        self.assertEqual(ai['available'], WANTED['available'])
        self.assertEqual(ai['allowed'], WANTED['allowed'])
        self.assertEqual(ai, WANTED)

    def test_create_from_ActionInformation(self):
        from Products.CMFCore.ActionInformation import ActionInformation

        WANTED = {'allowed': True, 'available': True, 'category': 'object',
                  'description': '', 'id': 'foo', 'title': 'foo', 'url': '',
                  'visible': True, 'icon': '', 'link_target': None}

        action = ActionInformation(id='foo')
        ec = None
        ai = self._makeOne(action, ec)

        self.assertEqual(ai['id'], WANTED['id'])
        self.assertEqual(ai['title'], WANTED['title'])
        self.assertEqual(ai['description'], WANTED['description'])
        self.assertEqual(ai['url'], WANTED['url'])
        self.assertEqual(ai['icon'], WANTED['icon'])
        self.assertEqual(ai['category'], WANTED['category'])
        self.assertEqual(ai['visible'], WANTED['visible'])
        self.assertEqual(ai['available'], WANTED['available'])
        self.assertEqual(ai['allowed'], WANTED['allowed'])
        self.assertEqual(ai, WANTED)

    def test_create_from_dict(self):
        WANTED = {'allowed': True, 'available': True, 'category': 'object',
                  'id': 'foo', 'title': 'foo', 'url': '', 'visible': True,
                  'icon': '', 'link_target': None}

        action = {'name': 'foo', 'url': ''}
        ec = None
        ai = self._makeOne(action, ec)

        self.assertEqual(ai['id'], WANTED['id'])
        self.assertEqual(ai['title'], WANTED['title'])
        self.assertEqual(ai['url'], WANTED['url'])
        self.assertEqual(ai['icon'], WANTED['icon'])
        self.assertEqual(ai['category'], WANTED['category'])
        self.assertEqual(ai['visible'], WANTED['visible'])
        self.assertEqual(ai['available'], WANTED['available'])
        self.assertEqual(ai['allowed'], WANTED['allowed'])
        self.assertEqual(ai, WANTED)


class ActionInfoSecurityTests(SecurityTest):

    def setUp(self):
        SecurityTest.setUp(self)
        self.site = DummySite('site').__of__(self.app)
        sm = getSiteManager()
        sm.registerUtility(DummyMembershipTool(), IMembershipTool)

    def tearDown(self):
        cleanUp()
        SecurityTest.tearDown(self)

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ActionInformation import ActionInfo

        return ActionInfo(*args, **kw)

    def test_create_from_dict(self):
        WANTED = {'allowed': True, 'available': True, 'category': 'object',
                  'id': 'foo', 'title': 'foo', 'url': '', 'visible': True,
                  'icon': '', 'link_target': None}

        action = {'name': 'foo', 'url': '', 'permissions': ('View',)}
        ec = createExprContext(self.site, self.site, None)
        ai = self._makeOne(action, ec)

        self.assertEqual(ai['id'], WANTED['id'])
        self.assertEqual(ai['title'], WANTED['title'])
        self.assertEqual(ai['url'], WANTED['url'])
        self.assertEqual(ai['icon'], WANTED['icon'])
        self.assertEqual(ai['category'], WANTED['category'])
        self.assertEqual(ai['visible'], WANTED['visible'])
        self.assertEqual(ai['available'], WANTED['available'])
        self.assertEqual(ai['allowed'], WANTED['allowed'])
        self.assertEqual(ai, WANTED)

    def test_category_object(self):
        # Permissions for action category 'object*' should be
        # evaluated in object context.
        manage_addFolder(self.site, 'actions_dummy')
        self.object = self.site.actions_dummy
        self.object.manage_permission('View', [], acquire=0)

        WANTED = {'allowed': False, 'category': 'object'}

        action = {'name': 'foo', 'url': '', 'permissions': ('View',)}
        ec = createExprContext(self.site, self.site, self.object)
        ai = self._makeOne(action, ec)

        self.assertEqual(ai['category'], WANTED['category'])
        self.assertEqual(ai['allowed'], WANTED['allowed'])

    def test_category_folder(self):
        # Permissions for action category 'folder*' should be
        # evaluated in folder context.
        manage_addFolder(self.site, 'actions_dummy')
        self.folder = self.site.actions_dummy
        self.folder.manage_permission('View', [], acquire=0)

        WANTED = {'allowed': False, 'category': 'folder'}

        action = {'name': 'foo', 'url': '', 'permissions': ('View',)}
        ec = createExprContext(self.folder, self.site, None)
        ai = self._makeOne(action, ec)
        ai['category'] = 'folder' # pfff

        self.assertEqual(ai['category'], WANTED['category'])
        self.assertEqual(ai['allowed'], WANTED['allowed'])

    def test_category_workflow(self):
        # Permissions for action category 'workflow*' should be
        # evaluated in object context.
        manage_addFolder(self.site, 'actions_dummy')
        self.object = self.site.actions_dummy
        self.object.manage_permission('View', [], acquire=0)

        WANTED = {'allowed': False, 'category': 'workflow'}

        action = {'name': 'foo', 'url': '', 'permissions': ('View',)}
        ec = createExprContext(self.site, self.site, self.object)
        ai = self._makeOne(action, ec)
        ai['category'] = 'workflow' # pfff

        self.assertEqual(ai['category'], WANTED['category'])
        self.assertEqual(ai['allowed'], WANTED['allowed'])

    def test_category_document(self):
        # Permissions for action category 'document*' should be
        # evaluated in object context (not in portal context).
        manage_addFolder(self.site, 'actions_dummy')
        self.object = self.site.actions_dummy
        self.object.manage_permission('View', [], acquire=0)

        WANTED = {'allowed': False, 'category': 'document'}

        action = {'name': 'foo', 'url': '', 'permissions': ('View',)}
        ec = createExprContext(self.site, self.site, self.object)
        ai = self._makeOne(action, ec)
        ai['category'] = 'document' # pfff

        self.assertEqual(ai['category'], WANTED['category'])
        self.assertEqual(ai['allowed'], WANTED['allowed'])

    def test_copy(self):
        action = {'name': 'foo', 'url': '', 'permissions': ('View',)}
        ec = createExprContext(self.site, self.site, None)
        ai = self._makeOne(action, ec)
        ai2 = ai.copy()

        self.assertEqual(ai._lazy_keys, ['allowed'])
        self.assertEqual(ai2._lazy_keys, ['allowed'])
        self.assertFalse(ai2._lazy_keys is ai._lazy_keys)
        self.assertEqual(ai['allowed'], True)
        self.assertEqual(ai2['allowed'], True)


class ActionInformationTests(TransactionalTest):

    layer = FunctionalZCMLLayer

    def setUp(self):
        TransactionalTest.setUp(self)

        self.app._setObject('portal', DummyContent('portal'))
        self.portal = self.app.portal
        sm = getSiteManager()
        sm.registerUtility(DummyMembershipTool(), IMembershipTool)
        self.folder = DummyContent('foo', 'url_foo')
        self.object = DummyContent('bar', 'url_bar')

    def tearDown(self):
        getSiteManager().unregisterUtility(provided=IMembershipTool)
        TransactionalTest.tearDown(self)

    def _makeOne(self, *args, **kw):
        from Products.CMFCore.ActionInformation import ActionInformation

        return ActionInformation(*args, **kw)

    def test_interfaces(self):
        from Products.CMFCore.ActionInformation import ActionInformation
        from Products.CMFCore.interfaces import IAction

        verifyClass(IAction, ActionInformation)

    def test_basic_construction(self):
        ai = self._makeOne(id='view')

        self.assertEqual(ai.getId(), 'view')
        self.assertEqual(ai.Title(), 'view')
        self.assertEqual(ai.Description(), '')
        self.assertEqual(ai.getCondition(), '')
        self.assertEqual(ai.getActionExpression(), '')
        self.assertEqual(ai.getVisibility(), 1)
        self.assertEqual(ai.getCategory(), 'object')
        self.assertEqual(ai.getPermissions(), ())

    def test_editing(self):
        ai = self._makeOne(id='view', category='folder')
        ai.edit(id='new_id', title='blah')

        self.assertEqual(ai.getId(), 'new_id')
        self.assertEqual(ai.Title(), 'blah')
        self.assertEqual(ai.Description(), '')
        self.assertEqual(ai.getCondition(), '')
        self.assertEqual(ai.getActionExpression(), '')
        self.assertEqual(ai.getVisibility(), 1)
        self.assertEqual(ai.getCategory(), 'folder')
        self.assertEqual(ai.getPermissions(), ())

    def test_setActionExpression_with_string_prefix(self):
        ai = self._makeOne(id='view', category='folder')
        ai.setActionExpression('string:blah')
        self.assertTrue(isinstance(ai.action, Expression))
        self.assertEqual(ai.getActionExpression(), 'string:blah')

    def test_construction_with_Expressions(self):
        ai = self._makeOne(id='view',
                           title='View',
                           action=Expression(text='view'),
                           condition=Expression(text='member'),
                           category='global',
                           visible=False)

        self.assertEqual(ai.getId(), 'view')
        self.assertEqual(ai.Title(), 'View')
        self.assertEqual(ai.Description(), '')
        self.assertEqual(ai.getCondition(), 'member')
        self.assertEqual(ai.getActionExpression(), 'string:${object_url}/view')
        self.assertEqual(ai.getVisibility(), 0)
        self.assertEqual(ai.getCategory(), 'global')
        self.assertEqual(ai.getPermissions(), ())

    def test_Condition(self):
        portal = self.portal
        folder = self.folder
        object = self.object
        ai = self._makeOne(id='view',
                           title='View',
                           action=Expression(text='view'),
                           condition=Expression(text='member'),
                           category='global',
                           visible=True)
        ec = createExprContext(folder, portal, object)

        self.assertFalse(ai.testCondition(ec))

    def test_Condition_PathExpression(self):
        portal = self.portal
        folder = self.folder
        object = self.object
        manage_addPythonScript(self.app, 'test_script')
        script = self.app.test_script
        script.ZPythonScript_edit('', 'return context.getId()')
        ai = self._makeOne(id='view',
                           title='View',
                           action=Expression(text='view'),
                           condition=Expression(text='portal/test_script'),
                           category='global',
                           visible=True)
        ec = createExprContext(folder, portal, object)

        self.assertTrue(ai.testCondition(ec))

    def test_getInfoData_empty(self):
        WANTED = ({'available': True, 'category': 'object',
                   'description': '', 'id': 'foo', 'permissions': (),
                   'title': 'foo', 'url': '', 'visible': True, 'icon': '',
                   'link_target': None}, [])
        a = self._makeOne('foo')
        self.assertEqual(a.getInfoData(), WANTED)

    def test_getInfoData_normal(self):
        a = self._makeOne('foo',
                          title='Foo Title',
                          description='Foo description.',
                          action='string:${object_url}/foo_url',
                          icon_expr='string:${object_url}/icon.gif',
                          condition='',
                          permissions=('View',),
                          visible=False,
                          link_target='_top')
        WANTED = ({'available': True, 'category': 'object',
                   'description': 'Foo description.', 'id': 'foo',
                   'permissions': ('View',), 'title': 'Foo Title',
                   'url': a._getActionObject(), 'visible': False,
                   'icon': a._getIconExpressionObject(),
                   'link_target': a.link_target},
                  ['url', 'icon'])
        self.assertEqual(a.getInfoData(), WANTED)


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(ActionCategoryTests),
        unittest.makeSuite(ActionTests),
        unittest.makeSuite(ActionInfoTests),
        unittest.makeSuite(ActionInfoSecurityTests),
        unittest.makeSuite(ActionInformationTests),
        ))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for security on FS* modules.
"""

import unittest
import Testing

import logging
from os import path

from AccessControl.Permission import Permission
from App.config import getConfiguration

from Products.CMFCore.tests.base.testcase import LogInterceptor
from Products.CMFCore.tests.base.testcase import WritableFSDVTest


class MetadataChecker(object):

    def _checkSettings(self, object, permissionname, acquire=0, roles=[]):
        # check the roles and acquire settings for a permission on an
        # object are as expected
        happy = 0
        for pstuff in object.ac_inherited_permissions(1):
            name, value = pstuff[:2]
            if name == permissionname:
                p = Permission(name, value, object)
                groles = p.getRoles(default=[])
                acquired = isinstance(groles, list)
                expected = {}
                for role in roles:
                    expected[role] = 1
                got = {}
                for role in groles:
                    got[role] = 1
                self.assertEqual((acquire, expected), (acquired, got))
                happy = 1
        if not happy:
            raise ValueError("'%s' not found in inherited permissions."
                             % permissionname)


class FSSecurityTests(WritableFSDVTest, MetadataChecker, LogInterceptor):

    def setUp(self):
        WritableFSDVTest.setUp(self)
        self._registerDirectory(self)

    def tearDown(self):
        self._ignore_log_errors(subsystem='CMFCore.DirectoryView')
        self._ignore_log_errors(subsystem='CMFCore.FSMetadata')
        WritableFSDVTest.tearDown(self)

    def test_basicPermissions(self):
        # Test basic FS permissions
        # check a normal method is as we'd expect
        self._checkSettings(self.ob.fake_skin.test1, 'View', 1, [])
        # now do some checks on the method with FS permissions
        self._checkSettings(self.ob.fake_skin.test4,
                            'View', 1, ['Manager', 'Owner'])
        self._checkSettings(self.ob.fake_skin.test4,
                            'View management screens', 1, [])

    def test_invalidPermissionNames(self):
        # Test for an invalid permission name
        from Products.CMFCore.DirectoryView import _dirreg

        self._catch_log_errors(logging.ERROR,
                               subsystem='CMFCore.DirectoryView')
        # baseline
        _dirreg.reloadDirectory(self.ob.fake_skin._dirpath)
        self._checkSettings(self.ob.fake_skin.test5, 'View', 1, [])
        self.assertEqual(self.logged, None)
        # add
        f = open(path.join(self.skin_path_name, 'test5.py.metadata'), 'w')
        f.write('[security]\nAccess stoopid contents = 0:')
        f.close()
        # test
        _dirreg.reloadDirectory(self.ob.fake_skin._dirpath)
        self._checkSettings(self.ob.fake_skin.test5, 'View', 1, [])
        self.assertEqual(len(self.logged), 1)
        self.assertEqual(self.logged[0].getMessage(),
                         'Error setting permissions')

    def test_invalidAcquireNames(self):
        # Test for an invalid spelling of acquire
        from Products.CMFCore.DirectoryView import _dirreg

        self._catch_log_errors(logging.ERROR,
                               subsystem='CMFCore.FSMetadata')
        # baseline
        _dirreg.reloadDirectory(self.ob.fake_skin._dirpath)
        self._checkSettings(self.ob.fake_skin.test5, 'View', 1, [])
        self.assertEqual(self.logged, None)
        # add
        f = open(path.join(self.skin_path_name, 'test5.py.metadata'), 'w')
        f.write('[security]\nView = aquire:')
        f.close()
        # test
        _dirreg.reloadDirectory(self.ob.fake_skin._dirpath)
        self._checkSettings(self.ob.fake_skin.test5, 'View', 1, [])
        self.assertEqual(len(self.logged), 1)
        self.assertEqual(self.logged[0].getMessage(),
                         'Error parsing .metadata file')


class DebugModeTests(WritableFSDVTest, MetadataChecker):

    def setUp(self):
        from Products.CMFCore.DirectoryView import _dirreg

        WritableFSDVTest.setUp(self)
        self._registerDirectory(self)
        info = _dirreg.getDirectoryInfo(self.ob.fake_skin._dirpath)
        self.use_dir_mtime = info.use_dir_mtime
        self.saved_cfg_debug_mode = getConfiguration().debug_mode
        getConfiguration().debug_mode = True

    def tearDown(self):
        getConfiguration().debug_mode = self.saved_cfg_debug_mode
        WritableFSDVTest.tearDown(self)

    def test_addPRM(self):
        # Test adding of a .metadata
        # baseline
        self._checkSettings(self.ob.fake_skin.test5, 'View', 1, [])
        # add
        self._writeFile('test5.py.metadata',
                        '[security]\nView = 1:Manager',
                        self.use_dir_mtime)
        # test
        self._checkSettings(self.ob.fake_skin.test5, 'View', 1, ['Manager'])

    def test_delPRM(self):
        # Test deleting of a .metadata
        # baseline
        self._checkSettings(self.ob.fake_skin.test4,
                            'View', 1, ['Manager', 'Owner'])
        # delete
        self._deleteFile('test4.py.metadata', self.use_dir_mtime)
        # test
        self._checkSettings(self.ob.fake_skin.test4, 'View', 1, [])

    def DISABLED_test_editPRM(self):
        # see https://bugs.launchpad.net/zope-cmf/+bug/714525
        # Test editing a .metadata
        # baseline
        self._checkSettings(self.ob.fake_skin.test4,
                            'View', 1, ['Manager', 'Owner'])
        # edit
        self._writeFile('test4.py.metadata',
                        '[security]\nView = 1:Manager',
                        self.use_dir_mtime)
        # test
        self._checkSettings(self.ob.fake_skin.test4, 'View', 1, ['Manager'])

    def DISABLED_test_DelAddEditPRM(self):
        # see https://bugs.launchpad.net/zope-cmf/+bug/714525
        # Test deleting, then adding, then editing a .metadata file
        # baseline
        self._checkSettings(self.ob.fake_skin.test4,
                            'View', 1, ['Manager', 'Owner'])
        # delete
        self._deleteFile('test4.py.metadata', self.use_dir_mtime)
        self._checkSettings(self.ob.fake_skin.test4, 'View', 1, [])
        # add back
        self._writeFile('test4.py.metadata',
                        '[security]\nView = 0:Manager,Anonymous',
                        self.use_dir_mtime)
        self._checkSettings(self.ob.fake_skin.test4,
                            'View', 0, ['Manager', 'Anonymous'])
        # edit
        self._writeFile('test4.py.metadata',
                        '[security]\nView = 1:Manager',
                        self.use_dir_mtime)
        # test
        self._checkSettings(self.ob.fake_skin.test4, 'View', 1, ['Manager'])


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(FSSecurityTests))
    suite.addTest(unittest.makeSuite(DebugModeTests))
    return suite



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Catalog tool setup handlers.
"""

from zope.component import getSiteManager

from Products.GenericSetup.utils import exportObjects
from Products.GenericSetup.utils import importObjects

from Products.CMFCore.interfaces import ICatalogTool


def importCatalogTool(context):
    """Import catalog tool.
    """
    sm = getSiteManager(context.getSite())
    tool = sm.queryUtility(ICatalogTool)
    if tool is None:
        logger = context.getLogger('catalog')
        logger.debug('Nothing to import.')
        return

    importObjects(tool, '', context)

def exportCatalogTool(context):
    """Export catalog tool.
    """
    sm = getSiteManager(context.getSite())
    tool = sm.queryUtility(ICatalogTool)
    if tool is None:
        logger = context.getLogger('catalog')
        logger.debug('Nothing to export.')
        return

    exportObjects(tool, '', context)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for FSDTMLMethod module.
"""

import unittest
import Testing

from os.path import join as path_join

from Acquisition import aq_base
from App.Common import rfc1123_date
from DateTime import DateTime
from OFS.Folder import Folder
from Products.StandardCacheManagers import RAMCacheManager
from zope.component import getSiteManager
from zope.component.hooks import setHooks
from zope.testing.cleanup import cleanUp

from Products.CMFCore.FSMetadata import FSMetadata
from Products.CMFCore.interfaces import ICachingPolicyManager
from Products.CMFCore.tests.base.dummy import DummyCachingManager
from Products.CMFCore.tests.base.dummy import DummyCachingManagerWithPolicy
from Products.CMFCore.tests.base.dummy import DummyContent
from Products.CMFCore.tests.base.testcase import FSDVTest
from Products.CMFCore.tests.base.testcase import SecurityTest
from Products.CMFCore.tests.base.testcase import TransactionalTest


class FSDTMLMaker(FSDVTest):

    def _makeOne(self, id, filename):
        from Products.CMFCore.FSDTMLMethod import FSDTMLMethod

        path = path_join(self.skin_path_name, filename)
        metadata = FSMetadata(path)
        metadata.read()
        return FSDTMLMethod(id, path, properties=metadata.getProperties())


class FSDTMLMethodTests(TransactionalTest, FSDTMLMaker):

    def setUp(self):
        TransactionalTest.setUp(self)
        FSDTMLMaker.setUp(self)
        setHooks()

    def tearDown(self):
        cleanUp()
        FSDTMLMaker.tearDown(self)
        TransactionalTest.tearDown(self)

    def test___call__(self):
        script = self._makeOne('testDTML', 'testDTML.dtml')
        script = script.__of__(self.app)
        self.assertEqual(script(self.app, self.REQUEST), 'nohost\n')

    def test_caching(self):
        #   Test HTTP caching headers.
        cpm = DummyCachingManager()
        getSiteManager().registerUtility(cpm, ICachingPolicyManager)
        original_len = len(self.RESPONSE.headers)
        obj = self._makeOne('testDTML', 'testDTML.dtml')
        obj = obj.__of__(self.app)
        obj(self.app, self.REQUEST, self.RESPONSE)
        self.assertTrue(len(self.RESPONSE.headers) >= original_len + 2)
        self.assertTrue('foo' in self.RESPONSE.headers.keys())
        self.assertTrue('bar' in self.RESPONSE.headers.keys())

    def test_ownership(self):
        script = self._makeOne('testDTML', 'testDTML.dtml')
        script = script.__of__(self.app)
        # fsdtmlmethod has no owner
        owner_tuple = script.getOwnerTuple()
        self.assertEqual(owner_tuple, None)

        # and ownership is not acquired [CMF/450]
        self.app._owner = ('/foobar', 'baz')
        owner_tuple = script.getOwnerTuple()
        self.assertEqual(owner_tuple, None)

    def test_304_response_from_cpm(self):
        # test that we get a 304 response from the cpm via this template
        mod_time = DateTime()
        cpm = DummyCachingManagerWithPolicy()
        getSiteManager().registerUtility(cpm, ICachingPolicyManager)
        content = DummyContent(id='content')
        content.modified_date = mod_time
        content = content.__of__(self.app)
        script = self._makeOne('testDTML', 'testDTML.dtml')
        script = script.__of__(content)
        self.REQUEST.environ['IF_MODIFIED_SINCE'
                            ] = '%s;' % rfc1123_date(mod_time + 3600)
        data = script(content, self.REQUEST, self.RESPONSE)

        self.assertEqual(data, '')
        self.assertEqual(self.RESPONSE.getStatus(), 304)


class FSDTMLMethodCustomizationTests(SecurityTest, FSDTMLMaker):

    def setUp(self):
        FSDTMLMaker.setUp(self)
        SecurityTest.setUp(self)
        self.skins, self.custom, self.fsdir, self.fsDTML = self._makeContext(
                                                   'testDTML', 'testDTML.dtml')

    def tearDown(self):
        cleanUp()
        SecurityTest.tearDown(self)
        FSDTMLMaker.tearDown(self)

    def test_customize(self):

        self.fsDTML.manage_doCustomize(folder_path='custom')

        self.assertEqual(len(self.custom.objectIds()), 1)
        self.assertTrue('testDTML' in self.custom.objectIds())

    def test_customize_alternate_root(self):
        self.app.other = Folder('other')

        self.fsDTML.manage_doCustomize(folder_path='other', root=self.app)

        self.assertFalse('testDTML' in self.custom.objectIds())
        self.assertTrue('testDTML' in self.app.other.objectIds())

    def test_customize_fspath_as_dot(self):

        self.fsDTML.manage_doCustomize(folder_path='.')

        self.assertFalse('testDTML' in self.custom.objectIds())
        self.assertTrue('testDTML' in self.skins.objectIds())

    def test_customize_manual_clone(self):
        clone = Folder('testDTML')

        self.fsDTML.manage_doCustomize(folder_path='custom', obj=clone)

        self.assertTrue('testDTML' in self.custom.objectIds())
        self.assertTrue(aq_base(self.custom._getOb('testDTML')) is clone)

    def test_customize_caching(self):
        # Test to ensure that cache manager associations survive customizing
        cache_id = 'gofast'
        RAMCacheManager.manage_addRAMCacheManager(self.app, cache_id,
                                                  REQUEST=None)
        self.fsDTML.ZCacheable_setManagerId(cache_id, REQUEST=None)

        self.assertEqual(self.fsDTML.ZCacheable_getManagerId(), cache_id)

        self.fsDTML.manage_doCustomize(folder_path='custom')
        custom_pt = self.custom.testDTML

        self.assertEqual(custom_pt.ZCacheable_getManagerId(), cache_id)


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(FSDTMLMethodTests),
        unittest.makeSuite(FSDTMLMethodCustomizationTests),
        ))



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Customizable DTML methods that come from the filesystem.
"""

import six
from six import get_unbound_function

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager
from App.special_dtml import DTMLFile
from App.special_dtml import HTML
from DocumentTemplate.security import RestrictedDTML
from OFS import bbb
from OFS.DTMLMethod import DTMLMethod, decapitate, guess_content_type
from OFS.role import RoleManager

from Products.CMFCore.DirectoryView import registerFileExtension
from Products.CMFCore.DirectoryView import registerMetaType
from Products.CMFCore.FSObject import FSObject
from Products.CMFCore.permissions import FTPAccess
from Products.CMFCore.permissions import View
from Products.CMFCore.permissions import ViewManagementScreens
from Products.CMFCore.utils import _checkConditionalGET
from Products.CMFCore.utils import _dtmldir
from Products.CMFCore.utils import _setCacheHeaders

_marker = object()


class FSDTMLMethod(RestrictedDTML, RoleManager, FSObject, HTML):

    """FSDTMLMethods act like DTML methods but are not directly
    modifiable from the management interface.
    """

    meta_type = 'Filesystem DTML Method'
    _owner = None
    _proxy_roles = ()
    _cache_namespace_keys = ()
    _reading = 0

    manage_options = (
        {'label': 'Customize', 'action': 'manage_main'},
        {'label': 'View', 'action': ''},
        {'label': 'Proxy', 'action': 'manage_proxyForm'})

    security = ClassSecurityInfo()
    security.declareObjectProtected(View)

    security.declareProtected(ViewManagementScreens, 'manage_main')
    manage_main = DTMLFile('custdtml', _dtmldir)

    def __init__(self, id, filepath, fullname=None, properties=None):
        FSObject.__init__(self, id, filepath, fullname, properties)
        # Normally called via HTML.__init__ but we don't need the rest that
        # happens there.
        self.initvars(None, {})

    def _createZODBClone(self):
        """Create a ZODB (editable) equivalent of this object."""
        return DTMLMethod(self.read(), __name__=self.getId())

    def _readFile(self, reparse):
        """Read the data from the filesystem.
        """
        file = open(self._filepath, 'r') # not 'rb', as this is a text file!
        try:
            data = file.read()
        finally:
            file.close()
        self.raw = data

        if reparse:
            self._reading = 1  # Avoid infinite recursion
            try:
                self.cook()
            finally:
                self._reading = 0

    # Hook up chances to reload in debug mode
    @security.private
    def read_raw(self):
        if not self._reading:
            self._updateFromFS()
        return HTML.read_raw(self)

    #### The following is mainly taken from OFS/DTMLMethod.py ###

    index_html = None # Prevent accidental acquisition

    # Documents masquerade as functions:
    __code__ = DTMLMethod.__code__

    default_content_type = 'text/html'

    def __call__(self, client=None, REQUEST={}, RESPONSE=None, **kw):
        """Render the document given a client object, REQUEST mapping,
        Response, and key word arguments."""

        self._updateFromFS()

        kw['document_id'] = self.getId()
        kw['document_title'] = self.title

        if client is not None:
            if _checkConditionalGET(self, kw):
                return ''

        if not self._cache_namespace_keys:
            data = self.ZCacheable_get(default=_marker)
            if data is not _marker:
                # Return cached results.
                return data

        __traceback_info__ = self._filepath
        security = getSecurityManager()
        security.addContext(self)
        try:
            r = HTML.__call__(self, client, REQUEST, **kw)

            if client is None:
                # Called as subtemplate, so don't need error propagation!
                if RESPONSE is None:
                    result = r
                else:
                    result = decapitate(r, RESPONSE)
                if not self._cache_namespace_keys:
                    self.ZCacheable_set(result)
                return result

            if not isinstance(r, six.string_types) or RESPONSE is None:
                if not self._cache_namespace_keys:
                    self.ZCacheable_set(r)
                return r

        finally:
            security.removeContext(self)

        headers = RESPONSE.headers
        if not ('content-type' in headers or 'Content-Type' in headers):
            if 'content_type' in self.__dict__:
                c = self.content_type
            else:
                if six.PY2:
                    c, _e = guess_content_type(self.getId(), r)
                else:
                    c, _e = guess_content_type(self.getId(), r.encode())

            RESPONSE.setHeader('Content-Type', c)
        if RESPONSE is not None:
            # caching policy manager hook
            _setCacheHeaders(self, {})
        result = decapitate(r, RESPONSE)
        if not self._cache_namespace_keys:
            self.ZCacheable_set(result)
        return result

    def getCacheNamespaceKeys(self):
        '''
        Returns the cacheNamespaceKeys.
        '''
        return self._cache_namespace_keys

    def setCacheNamespaceKeys(self, keys, REQUEST=None):
        '''
        Sets the list of names that should be looked up in the
        namespace to provide a cache key.
        '''
        ks = []
        for key in keys:
            key = str(key).strip()
            if key:
                ks.append(key)
        self._cache_namespace_keys = tuple(ks)
        if REQUEST is not None:
            return self.ZCacheable_manage(self, REQUEST)

    # Zope 2.3.x way:
    def validate(self, inst, parent, name, value, md=None):
        return getSecurityManager().validate(inst, parent, name, value)

    if bbb.HAS_ZSERVER:
        security.declareProtected(FTPAccess, 'manage_FTPget')
        manage_FTPget = get_unbound_function(DTMLMethod.manage_FTPget)

    security.declareProtected(ViewManagementScreens, 'PrincipiaSearchSource')
    PrincipiaSearchSource = get_unbound_function(DTMLMethod.PrincipiaSearchSource)

    security.declareProtected(ViewManagementScreens, 'document_src')
    document_src = get_unbound_function(DTMLMethod.document_src)

    security.declareProtected(ViewManagementScreens, 'manage_haveProxy')
    manage_haveProxy = get_unbound_function(DTMLMethod.manage_haveProxy)

InitializeClass(FSDTMLMethod)

registerFileExtension('dtml', FSDTMLMethod)
registerMetaType('DTML Method', FSDTMLMethod)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Basic undo tool.
"""

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from AccessControl.SecurityManagement import getSecurityManager
from App.special_dtml import DTMLFile
from OFS.SimpleItem import SimpleItem
from zope.interface import implementer

from Products.CMFCore.exceptions import AccessControl_Unauthorized
from Products.CMFCore.interfaces import IUndoTool
from Products.CMFCore.permissions import ListUndoableChanges
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.utils import _checkPermission
from Products.CMFCore.utils import _dtmldir
from Products.CMFCore.utils import registerToolInterface
from Products.CMFCore.utils import UniqueObject


@implementer(IUndoTool)
class UndoTool(UniqueObject, SimpleItem):

    """ This tool is used to undo changes.
    """

    id = 'portal_undo'
    meta_type = 'CMF Undo Tool'

    security = ClassSecurityInfo()

    manage_options = ( SimpleItem.manage_options
                     + ({'label': 'Overview',
                         'action': 'manage_overview'},)
                     )

    #
    #   ZMI methods
    #
    security.declareProtected(ManagePortal, 'manage_overview')
    manage_overview = DTMLFile( 'explainUndoTool', _dtmldir )

    #
    #   'IUndoTool' interface methods
    #
    @security.protected(ListUndoableChanges)
    def listUndoableTransactionsFor(self, object,
                                    first_transaction=None,
                                    last_transaction=None,
                                    PrincipiaUndoBatchSize=None):
        """ List all transaction IDs the user is allowed to undo on 'object'.
        """
        transactions = object.undoable_transactions(
            first_transaction=first_transaction,
            last_transaction=last_transaction,
            PrincipiaUndoBatchSize=PrincipiaUndoBatchSize)
        for t in transactions:
            # Ensure transaction ids don't have embedded LF.
            t['id'] = t['id'].replace('\n', '')
        if not _checkPermission(ManagePortal, object):
            # Filter out transactions done by other members of the portal.
            user_id = getSecurityManager().getUser().getId()
            transactions = tuple(filter(
                lambda record, user_id=user_id:
                record['user_name'].split()[-1] == user_id,
                transactions
                ))
        return transactions

    security.declarePublic('undo')
    def undo(self, object, transaction_info):
        """
            Undo the list of transactions passed in 'transaction_info',
            first verifying that the current user is allowed to undo them.
        """
        # Belt and suspenders:  make sure that the user is actually
        # allowed to undo the transation(s) in transaction_info.

        xids = {}  # set of allowed transaction IDs

        allowed = self.listUndoableTransactionsFor( object )

        for xid in [x['id'] for x in allowed]:
            xids[xid] = 1

        if type( transaction_info ) == type( '' ):
            transaction_info = [ transaction_info ]

        for tinfo in transaction_info:
            if not xids.get( tinfo, None ):
                raise AccessControl_Unauthorized

        object.manage_undo_transactions(transaction_info)

InitializeClass(UndoTool)
registerToolInterface('portal_undo', IUndoTool)



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Member data tool xml adapter and setup handler unit tests.
"""

import unittest

from DateTime.DateTime import DateTime
from OFS.Folder import Folder
from zope.component import getSiteManager

from Products.GenericSetup.testing import BodyAdapterTestCase
from Products.GenericSetup.tests.common import BaseRegistryTests
from Products.GenericSetup.tests.common import DummyExportContext
from Products.GenericSetup.tests.common import DummyImportContext

from Products.CMFCore.interfaces import IMemberDataTool
from Products.CMFCore.MemberDataTool import MemberDataTool
from Products.CMFCore.testing import ExportImportZCMLLayer

_MEMBERDATATOOL_BODY = b"""\
<?xml version="1.0" encoding="utf-8"?>
<object name="portal_memberdata" meta_type="CMF Member Data Tool">
 <property name="email"></property>
 <property name="portal_skin"></property>
 <property name="listed">False</property>
 <property name="login_time">1970/01/01 00:00:00 UTC</property>
 <property name="last_login_time">1970/01/01 00:00:00 UTC</property>
 <property name="fullname"></property>
</object>
"""

_DEFAULT_EXPORT = b"""\
<?xml version="1.0" encoding="utf-8"?>
<object name="portal_memberdata" meta_type="CMF Member Data Tool">
 <property name="email"></property>
 <property name="portal_skin"></property>
 <property name="listed">False</property>
 <property name="login_time">1970/01/01 00:00:00 UTC</property>
 <property name="last_login_time">1970/01/01 00:00:00 UTC</property>
 <property name="fullname"></property>
</object>
"""

_CHANGED_EXPORT = """\
<?xml version="1.0"?>
<object name="portal_memberdata" meta_type="CMF Member Data Tool">
 <property name="email">value1</property>
 <property name="portal_skin">value2</property>
 <property name="listed">True</property>
 <property name="login_time">2010/01/01 00:00:00</property>
 <property name="last_login_time">2010/01/01 00:00:00</property>
 <property name="fullname">value3</property>
 <property name="home_page" type="string"></property>
</object>
"""


class MemberDataToolXMLAdapterTests(BodyAdapterTestCase, unittest.TestCase):

    layer = ExportImportZCMLLayer

    def _getTargetClass(self):
        from Products.CMFCore.exportimport.memberdata \
                import MemberDataToolXMLAdapter

        return MemberDataToolXMLAdapter

    def setUp(self):
        self._obj = MemberDataTool()
        self._BODY = _MEMBERDATATOOL_BODY


class _MemberDataToolSetup(BaseRegistryTests):

    def _initSite(self, use_changed=False):
        site = Folder(id='site').__of__(self.app)
        mdtool = MemberDataTool()
        getSiteManager().registerUtility(mdtool, IMemberDataTool)

        if use_changed:
            mdtool._updateProperty('email', 'value1')
            mdtool._updateProperty('portal_skin', 'value2')
            mdtool._updateProperty('listed', 'True')
            mdtool._updateProperty('login_time', '2010/01/01')
            mdtool._updateProperty('last_login_time', '2010/01/01')
            mdtool._updateProperty('fullname', 'value3')
            mdtool.manage_addProperty('home_page', '', 'string')

        return site, mdtool


class exportMemberDataToolTests(_MemberDataToolSetup):

    layer = ExportImportZCMLLayer

    def test_unchanged(self):
        from Products.CMFCore.exportimport.memberdata \
                import exportMemberDataTool

        site, _mdtool = self._initSite(use_changed=False)
        context = DummyExportContext(site)
        exportMemberDataTool(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'memberdata.xml')
        self._compareDOM(text.decode('utf8'), _DEFAULT_EXPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_changed(self):
        from Products.CMFCore.exportimport.memberdata \
                import exportMemberDataTool

        site, _mdtool = self._initSite(use_changed=True)
        context = DummyExportContext(site)
        exportMemberDataTool(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'memberdata.xml')
        self._compareDOM(text.decode('utf8'), _CHANGED_EXPORT)
        self.assertEqual(content_type, 'text/xml')


class importMemberDataToolTests(_MemberDataToolSetup):

    layer = ExportImportZCMLLayer

    def test_normal(self):
        from Products.CMFCore.exportimport.memberdata \
                import importMemberDataTool

        site, mdtool = self._initSite()

        context = DummyImportContext(site)
        context._files['memberdata.xml'] = _CHANGED_EXPORT
        importMemberDataTool(context)

        self.assertEqual(mdtool.email, 'value1')
        self.assertEqual(mdtool.portal_skin, 'value2')
        self.assertEqual(mdtool.listed, True)
        self.assertEqual(mdtool.login_time, DateTime('2010/01/01'))
        self.assertEqual(mdtool.last_login_time, DateTime('2010/01/01'))


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(MemberDataToolXMLAdapterTests),
        unittest.makeSuite(exportMemberDataToolTests),
        unittest.makeSuite(importMemberDataToolTests),
        ))



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Member data tool xml adapter and setup handlers.
"""

from zope.component import adapts
from zope.component import getSiteManager

from Products.GenericSetup.interfaces import ISetupEnviron
from Products.GenericSetup.utils import exportObjects
from Products.GenericSetup.utils import importObjects
from Products.GenericSetup.utils import PropertyManagerHelpers
from Products.GenericSetup.utils import XMLAdapterBase

from Products.CMFCore.interfaces import IMemberDataTool


class MemberDataToolXMLAdapter(XMLAdapterBase, PropertyManagerHelpers):

    """XML im- and exporter for MemberDataTool.
    """

    adapts(IMemberDataTool, ISetupEnviron)

    _LOGGER_ID = 'memberdata'

    name = 'memberdata'

    def _exportNode(self):
        """Export the object as a DOM node.
        """
        node = self._getObjectNode('object')
        node.appendChild(self._extractProperties())

        self._logger.info('Member data tool exported.')
        return node

    def _importNode(self, node):
        """Import the object from the DOM node.
        """
        if self.environ.shouldPurge():
            self._purgeProperties()

        self._initProperties(node)

        self._logger.info('Member data tool imported.')


def importMemberDataTool(context):
    """Import member data tool settings from an XML file.
    """
    sm = getSiteManager(context.getSite())
    tool = sm.queryUtility(IMemberDataTool)
    if tool is None:
        logger = context.getLogger('memberdata')
        logger.debug('Nothing to import.')
        return

    importObjects(tool, '', context)

def exportMemberDataTool(context):
    """Export member data tool settings as an XML file.
    """
    sm = getSiteManager(context.getSite())
    tool = sm.queryUtility(IMemberDataTool)
    if tool is None:
        logger = context.getLogger('memberdata')
        logger.debug('Nothing to export.')
        return

    exportObjects(tool, '', context)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Cookie Crumbler: Enable cookies for non-cookie user folders.
"""

import base64
import six
from six.moves.urllib.parse import quote, unquote

from AccessControl.class_init import InitializeClass
from AccessControl.Permissions import view_management_screens
from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import aq_inner
from Acquisition import aq_parent
from App.special_dtml import HTMLFile
from DateTime.DateTime import DateTime
from OFS.interfaces import IObjectWillBeMovedEvent
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from zope.component import getUtility
from zope.container.interfaces import IObjectMovedEvent
from zope.globalrequest import getRequest
from zope.interface import implementer
from ZPublisher import BeforeTraverse
from ZPublisher.HTTPRequest import HTTPRequest

from Products.CMFCore.interfaces import IActionsTool
from Products.CMFCore.interfaces import ICookieCrumbler
from Products.CMFCore.utils import registerToolInterface
from Products.CMFCore.utils import UniqueObject


# Constants.
ATTEMPT_NONE = 0       # No attempt at authentication
ATTEMPT_LOGIN = 1      # Attempt to log in
ATTEMPT_RESUME = 2     # Attempt to resume session

ModifyCookieCrumblers = 'Modify Cookie Crumblers'
ViewManagementScreens = view_management_screens


class CookieCrumblerDisabled(Exception):

    """Cookie crumbler should not be used for a certain request.
    """


@implementer(ICookieCrumbler)
class CookieCrumbler(UniqueObject, PropertyManager, SimpleItem):

    """Reads cookies during traversal and simulates the HTTP auth headers.
    """

    id = 'cookie_authentication'

    manage_options = (
        PropertyManager.manage_options +
        SimpleItem.manage_options)

    meta_type = 'Cookie Crumbler'

    security = ClassSecurityInfo()
    security.declareProtected(ModifyCookieCrumblers, 'manage_editProperties')
    security.declareProtected(ModifyCookieCrumblers, 'manage_changeProperties')
    security.declareProtected(ViewManagementScreens, 'manage_propertiesForm')

    # By default, anonymous users can view login/logout pages.
    _View_Permission = ('Anonymous',)

    _properties = ({'id': 'auth_cookie', 'type': 'string', 'mode': 'w',
                    'label': 'Authentication cookie name'},
                   {'id': 'name_cookie', 'type': 'string', 'mode': 'w',
                    'label': 'User name form variable'},
                   {'id': 'pw_cookie', 'type': 'string', 'mode': 'w',
                    'label': 'User password form variable'},
                   {'id': 'persist_cookie', 'type': 'string', 'mode': 'w',
                    'label': 'User name persistence form variable'},
                   {'id': 'local_cookie_path', 'type': 'boolean', 'mode': 'w',
                    'label': 'Use cookie paths to limit scope'},
                   {'id': 'cache_header_value', 'type': 'string', 'mode': 'w',
                    'label': 'Cache-Control header value'},
                   {'id': 'log_username', 'type': 'boolean', 'mode': 'w',
                    'label': 'Log cookie auth username to access log'}
                   )

    auth_cookie = '__ac'
    name_cookie = '__ac_name'
    pw_cookie = '__ac_password' # not used as cookie, just as request key
    persist_cookie = '__ac_persistent'
    local_cookie_path = False
    cache_header_value = 'private'
    log_username = True

    @security.private
    def delRequestVar(self, req, name):
        # No errors of any sort may propagate, and we don't care *what*
        # they are, even to log them.
        try:
            del req.other[name]
        except:
            pass
        try:
            del req.form[name]
        except:
            pass
        try:
            del req.cookies[name]
        except:
            pass
        try:
            del req.environ[name]
        except:
            pass

    @security.public
    def getCookiePath(self):
        if not self.local_cookie_path:
            return '/'
        parent = aq_parent(aq_inner(self))
        if parent is not None:
            return '/' + parent.absolute_url(1)
        else:
            return '/'

    # Allow overridable cookie set/expiration methods.
    @security.private
    def getCookieMethod(self, name, default=None):
        return getattr(self, name, default)

    @security.private
    def defaultSetAuthCookie(self, resp, cookie_name, cookie_value):
        kw = {}
        req = getRequest()
        if req is not None and req.get('SERVER_URL', '').startswith('https:'):
            # Ask the client to send back the cookie only in SSL mode
            kw['secure'] = 'y'
        resp.setCookie(cookie_name, cookie_value,
                       path=self.getCookiePath(), **kw)

    @security.private
    def defaultExpireAuthCookie(self, resp, cookie_name):
        resp.expireCookie(cookie_name, path=self.getCookiePath())

    def _setAuthHeader(self, ac, request, response):
        """Set the auth headers for both the Zope and Medusa http request
        objects.
        """
        request._auth = 'Basic %s' % ac
        response._auth = 1
        if self.log_username:
            # Set the authorization header in the medusa http request
            # so that the username can be logged to the Z2.log
            try:
                # Put the full-arm latex glove on now...
                medusa_headers = response.stdout._request._header_cache
            except AttributeError:
                pass
            else:
                medusa_headers['authorization'] = request._auth

    @security.private
    def modifyRequest(self, req, resp):
        """Copies cookie-supplied credentials to the basic auth fields.

        Returns a flag indicating what the user is trying to do with
        cookies: ATTEMPT_NONE, ATTEMPT_LOGIN, or ATTEMPT_RESUME.  If
        cookie login is disabled for this request, raises
        CookieCrumblerDisabled.
        """
        if (not isinstance(req, HTTPRequest)
            or not req['REQUEST_METHOD'] in ('HEAD', 'GET', 'PUT', 'POST')
            or 'WEBDAV_SOURCE_PORT' in req.environ):
            raise CookieCrumblerDisabled

        # attempt may contain information about an earlier attempt to
        # authenticate using a higher-up cookie crumbler within the
        # same request.
        attempt = getattr(req, '_cookie_auth', ATTEMPT_NONE)

        if attempt == ATTEMPT_NONE:
            if req._auth:
                # An auth header was provided and no cookie crumbler
                # created it.  The user must be using basic auth.
                raise CookieCrumblerDisabled

            if self.pw_cookie in req and self.name_cookie in req:
                # Attempt to log in and set cookies.
                attempt = ATTEMPT_LOGIN
                name = req[self.name_cookie]
                pw = req[self.pw_cookie]
                if six.PY2:
                    ac = base64.encodestring('%s:%s' % (name, pw)).rstrip()
                else:
                    ac = base64.encodebytes(
                        ('%s:%s' % (name, pw)).encode()
                    ).rstrip().decode()
                self._setAuthHeader(ac, req, resp)
                if req.get(self.persist_cookie, 0):
                    # Persist the user name (but not the pw or session)
                    expires = (DateTime() + 365).toZone('GMT').rfc822()
                    resp.setCookie(self.name_cookie, name,
                                   path=self.getCookiePath(),
                                   expires=expires)
                else:
                    # Expire the user name
                    resp.expireCookie(self.name_cookie,
                                      path=self.getCookiePath())
                method = self.getCookieMethod('setAuthCookie',
                                              self.defaultSetAuthCookie)
                method(resp, self.auth_cookie, quote(ac))
                self.delRequestVar(req, self.name_cookie)
                self.delRequestVar(req, self.pw_cookie)

            elif self.auth_cookie in req:
                # Attempt to resume a session if the cookie is valid.
                # Copy __ac to the auth header.
                ac = unquote(req[self.auth_cookie])
                if ac and ac != 'deleted':
                    try:
                        if six.PY2:
                            base64.decodestring(ac)
                        else:
                            base64.decodebytes(ac.encode())
                    except:
                        # Not a valid auth header.
                        pass
                    else:
                        attempt = ATTEMPT_RESUME
                        self._setAuthHeader(ac, req, resp)
                        self.delRequestVar(req, self.auth_cookie)
                        method = self.getCookieMethod(
                            'twiddleAuthCookie', None)
                        if method is not None:
                            method(resp, self.auth_cookie, quote(ac))

        req._cookie_auth = attempt
        return attempt

    def __call__(self, container, req):
        '''The __before_publishing_traverse__ hook.'''
        resp = req['RESPONSE']
        try:
            attempt = self.modifyRequest(req, resp)
        except CookieCrumblerDisabled:
            return
        if attempt != ATTEMPT_NONE:
            # Trying to log in or resume a session
            if self.cache_header_value:
                # we don't want caches to cache the resulting page
                resp.setHeader('Cache-Control', self.cache_header_value)
                # demystify this in the response.
                resp.setHeader('X-Cache-Control-Hdr-Modified-By',
                               'CookieCrumbler')
            phys_path = self.getPhysicalPath()
            # Cookies are in use.
            # Provide a logout page.
            req._logout_path = phys_path + ('logout',)

    @security.public
    def credentialsChanged(self, user, name, pw, request=None):
        """
        Updates cookie credentials if user details are changed.
        """
        if request is None:
            request = getRequest() # BBB for Membershiptool
        reponse = request['RESPONSE']
        ac = base64.encodestring('%s:%s' % (name, pw)).rstrip()
        method = self.getCookieMethod('setAuthCookie',
                                       self.defaultSetAuthCookie)
        method(reponse, self.auth_cookie, quote(ac))

    @security.public
    def logout(self, response=None):
        """
        Logs out the user
        """
        target = None
        if response is None:
            response = getRequest()['RESPONSE'] # BBB for App.Management
            atool = getUtility(IActionsTool)
            target = atool.getActionInfo('user/logout')['url']
        method = self.getCookieMethod('expireAuthCookie',
                                       self.defaultExpireAuthCookie)
        method(response, cookie_name=self.auth_cookie)
        # BBB for App.Management
        if target is not None:
            response.redirect(target)

    @security.public
    def propertyLabel(self, id):
        """Return a label for the given property id
        """
        for p in self._properties:
            if p['id'] == id:
                return p.get('label', id)
        return id

InitializeClass(CookieCrumbler)
registerToolInterface('cookie_authentication', ICookieCrumbler)


def handleCookieCrumblerEvent(ob, event):
    """ Event subscriber for (un)registering a CC as a before traverse hook.
    """
    if not ICookieCrumbler.providedBy(ob):
        return

    if IObjectMovedEvent.providedBy(event):
        if event.newParent is not None:
            # register before traverse hook
            handle = ob.meta_type + '/' + ob.getId()
            nc = BeforeTraverse.NameCaller(ob.getId())
            BeforeTraverse.registerBeforeTraverse(event.newParent, nc, handle)
    elif IObjectWillBeMovedEvent.providedBy(event):
        if event.oldParent is not None:
            # unregister before traverse hook
            handle = ob.meta_type + '/' + ob.getId()
            BeforeTraverse.unregisterBeforeTraverse(event.oldParent, handle)


manage_addCCForm = HTMLFile('dtml/addCC', globals())
manage_addCCForm.__name__ = 'addCC'

def manage_addCC(dispatcher, id, REQUEST=None):
    ' '
    ob = CookieCrumbler()
    ob.id = id
    dispatcher._setObject(ob.getId(), ob)
    ob = getattr(dispatcher.this(), ob.getId())
    if REQUEST is not None:
        return dispatcher.manage_main(dispatcher, REQUEST)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for FSPythonScript module.
"""

import unittest
from Testing import ZopeTestCase
ZopeTestCase.installProduct('PythonScripts', 1)

import os
import warnings
from os.path import join
from six.moves._thread import start_new_thread
from sys import exc_info
from time import sleep

from Acquisition import aq_base
from DateTime.DateTime import DateTime
from OFS.Folder import Folder
from OFS.SimpleItem import SimpleItem
from Products.StandardCacheManagers import RAMCacheManager
from zope.testing.cleanup import cleanUp

from Products.CMFCore.FSMetadata import FSMetadata
from Products.CMFCore.FSPythonScript import FSPythonScript
from Products.CMFCore.tests.base.testcase import FSDVTest
from Products.CMFCore.tests.base.testcase import SecurityTest


class FSPSMaker(FSDVTest):

    def _makeOne(self, id, filename):
        path = join(self.skin_path_name, filename)
        metadata = FSMetadata(path)
        metadata.read()
        return FSPythonScript(id, path, properties=metadata.getProperties())


class FSPythonScriptTests(FSPSMaker):

    def test_get_size(self):
        # Test get_size returns correct value
        script = self._makeOne('test1', 'test1.py')
        self.assertEqual(len(script.read()), script.get_size())

    def test_getModTime(self):
        script = self._makeOne('test1', 'test1.py')
        self.assertTrue(isinstance(script.getModTime(), DateTime))
        self.assertEqual(script.getModTime(),
                         DateTime(os.stat(script._filepath).st_mtime))

    def test_bobobase_modification_time(self):
        script = self._makeOne('test1', 'test1.py')
        self.assertTrue(isinstance(script.bobobase_modification_time(),
                                   DateTime))
        self.assertEqual(script.bobobase_modification_time(),
                         DateTime(os.stat(script._filepath).st_mtime))

    def test_initialization_race_condition(self):
        # Tries to exercise a former race condition where
        # FSObject._updateFromFS() set self._parsed before the
        # object was really parsed.
        for _n in range(10):
            f = Folder()
            script = self._makeOne('test1', 'test1.py').__of__(f)
            res = []

            def call_script(script=script, res=res):
                try:
                    res.append(script())
                except:
                    res.append('%s: %s' % exc_info()[:2])

            start_new_thread(call_script, ())
            call_script()
            while len(res) < 2:
                sleep(0.05)
            self.assertEqual(res, ['test1', 'test1'], res)

    def test_foreign_line_endings(self):
        # Load the various line ending files and get their output
        container = Folder('container_for_execution')
        for fformat in ('unix', 'dos', 'mac'):
            container._setObject(fformat,
                self._makeOne(fformat, 'test_%s.py' % fformat))
            script = getattr(container, fformat)
            self.assertEqual(script(), fformat)


class FSPythonScriptCustomizationTests(SecurityTest, FSPSMaker):

    def setUp(self):
        FSPSMaker.setUp(self)
        SecurityTest.setUp(self)

    def tearDown(self):
        cleanUp()
        SecurityTest.tearDown(self)
        FSPSMaker.tearDown(self)

    def test_customize(self):
        from Products.CMFCore.FSPythonScript import CustomizedPythonScript

        _stool, custom, _fsdir, fsPS = self._makeContext('test6', 'test6.py')

        fsPS.manage_doCustomize(folder_path='custom')

        self.assertEqual(len(custom.objectIds()), 1)
        self.assertTrue('test6' in custom.objectIds())

        test6 = custom._getOb('test6')

        self.assertTrue(isinstance(test6, CustomizedPythonScript))
        self.assertEqual(test6.original_source, fsPS.read())

    def test_customize_alternate_root(self):
        _stool, custom, _fsdir, fsPS = self._makeContext('test6', 'test6.py')
        self.app.other = Folder('other')

        fsPS.manage_doCustomize(folder_path='other', root=self.app)

        self.assertFalse('test6' in custom.objectIds())
        self.assertTrue('test6' in self.app.other.objectIds())

    def test_customize_fspath_as_dot(self):
        stool, custom, _fsdir, fsPS = self._makeContext('test6', 'test6.py')

        fsPS.manage_doCustomize(folder_path='.')

        self.assertFalse('test6' in custom.objectIds())
        self.assertTrue('test6' in stool.objectIds())

    def test_customize_manual_clone(self):
        _stool, custom, _fsdir, fsPS = self._makeContext('test6', 'test6.py')
        clone = Folder('test6')

        fsPS.manage_doCustomize(folder_path='custom', obj=clone)

        self.assertTrue('test6' in custom.objectIds())
        self.assertTrue(aq_base(custom._getOb('test6')) is clone)

    def test_customize_caching(self):
        # Test to ensure that cache manager associations survive customizing
        _stool, custom, _fsdir, fsPS = self._makeContext('test6', 'test6.py')

        cache_id = 'gofast'
        RAMCacheManager.manage_addRAMCacheManager(self.app, cache_id,
                                                  REQUEST=None)
        fsPS.ZCacheable_setManagerId(cache_id, REQUEST=None)

        self.assertEqual(fsPS.ZCacheable_getManagerId(), cache_id)

        fsPS.manage_doCustomize(folder_path='custom')
        custom_ps = custom.test6

        self.assertEqual(custom_ps.ZCacheable_getManagerId(), cache_id)

    def test_customize_proxyroles(self):
        # Test to ensure that proxy roles survive customizing
        _stool, custom, _fsdir, fsPS = self._makeContext('test6', 'test6.py')

        fsPS._proxy_roles = ('Manager', 'Anonymous')
        self.assertTrue(fsPS.manage_haveProxy('Anonymous'))
        self.assertTrue(fsPS.manage_haveProxy('Manager'))

        fsPS.manage_doCustomize(folder_path='custom')
        custom_ps = custom.test6
        self.assertTrue(custom_ps.manage_haveProxy('Anonymous'))
        self.assertTrue(custom_ps.manage_haveProxy('Manager'))

    def test_customization_permissions(self):
        # Test to ensure that permission settings survive customizing
        _stool, custom, _fsdir, fsPS = self._makeContext('test6', 'test6.py')
        perm = 'View management screens'

        # First, set a permission to an odd role and verify
        fsPS.manage_permission(perm, roles=('Anonymous',), acquire=0)
        rop = fsPS.rolesOfPermission(perm)
        for rop_info in rop:
            if rop_info['name'] == 'Anonymous':
                self.assertFalse(rop_info['selected'] == '')
            else:
                self.assertTrue(rop_info['selected'] == '')

        # Now customize and verify again
        fsPS.manage_doCustomize(folder_path='custom')
        custom_ps = custom.test6
        rop = custom_ps.rolesOfPermission(perm)
        for rop_info in rop:
            if rop_info['name'] == 'Anonymous':
                self.assertFalse(rop_info['selected'] == '')
            else:
                self.assertTrue(rop_info['selected'] == '')

_ORIGINAL_TEXT = """\
## Script (Python) "cps"
##bind container=container
##bind context=context
##bind namespace=
##bind script=script
##bind subpath=traverse_subpath
##parameters=
##title=
##
return 'cps'
"""

_REPLACEMENT_TEXT = """\
## Script (Python) "cps"
##bind container=container
##bind context=context
##bind namespace=
##bind script=script
##bind subpath=traverse_subpath
##parameters=
##title=
##
return 'cps -- replaced'
"""

_DIFF_TEXT = """\
--- original
+++ modified
@@ -7,4 +7,4 @@
 ##parameters=
 ##title=
 ##
-return 'cps'
+return 'cps -- replaced'
"""


class CustomizedPythonScriptTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.CMFCore.FSPythonScript import CustomizedPythonScript
        return CustomizedPythonScript

    def _makeOne(self, id, text):
        return self._getTargetClass()(id, text)

    def test_write_leaves_original_source(self):
        cps = self._makeOne('cps', _ORIGINAL_TEXT)
        self.assertEqual(cps.read(), _ORIGINAL_TEXT)
        self.assertEqual(cps.original_source, _ORIGINAL_TEXT)
        cps.write(_REPLACEMENT_TEXT)
        self.assertEqual(cps.read(), _REPLACEMENT_TEXT)
        self.assertEqual(cps.original_source, _ORIGINAL_TEXT)

    def test_getDiff(self):
        cps = self._makeOne('cps', _ORIGINAL_TEXT)
        self.assertEqual(len(list(cps.getDiff())), 0)

        cps.write(_REPLACEMENT_TEXT)
        self.assertEqual([l.rstrip() for l in cps.getDiff()],
                         _DIFF_TEXT.splitlines())


class WarnMe(SimpleItem):
    """Emits a UserWarning when called"""

    def __init__(self, stacklevel):
        self._stacklevel = stacklevel

    def __call__(self):
        warnings.warn('foo', stacklevel=self._stacklevel)


class FSPythonScriptWarningsTests(SecurityTest, FSPSMaker):

    def setUp(self):
        SecurityTest.setUp(self)
        FSPSMaker.setUp(self)

    def tearDown(self):
        FSPSMaker.tearDown(self)
        SecurityTest.tearDown(self)

    def testFSPSWarn(self):
        self.app._setObject('warn_me', WarnMe(2))
        self.app._setObject('warn1', self._makeOne('warn1', 'test_warn.py'))
        # This used to raise an error:
        #   File "/usr/local/python2.3/lib/python2.3/warnings.py", line 63,
        #   in warn_explicit
        #     if module[-3:].lower() == ".py":
        # TypeError: unsubscriptable object
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self.app.warn_me()
            self.app.warn1()


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(FSPythonScriptTests),
        unittest.makeSuite(FSPythonScriptCustomizationTests),
        unittest.makeSuite(CustomizedPythonScriptTests),
        unittest.makeSuite(FSPythonScriptWarningsTests),
        ))



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Common pieces of the workflow architecture.
"""

from zope.interface import implementer
from zope.component.interfaces import ObjectEvent

from Products.CMFCore.interfaces import IWorkflowActionEvent
from Products.CMFCore.interfaces import IActionWillBeInvokedEvent
from Products.CMFCore.interfaces import IActionRaisedExceptionEvent
from Products.CMFCore.interfaces import IActionSucceededEvent

class WorkflowException( Exception ):

    """ Exception while invoking workflow.
    """
    

class ObjectDeleted( Exception ):

    """ Raise to tell the workflow tool that the object has been deleted.

    Swallowed by the workflow tool.
    """
    def __init__(self, result=None):
        self._r = result

    def getResult(self):
        return self._r


class ObjectMoved( Exception ):

    """ Raise to tell the workflow tool that the object has moved.

    Swallowed by the workflow tool.
    """
    def __init__(self, new_ob, result=None):
        self._ob = new_ob  # Includes acquisition wrappers.
        self._r = result

    def getResult(self):
        return self._r

    def getNewObject(self):
        return self._ob

# Events

@implementer(IWorkflowActionEvent)
class WorkflowActionEvent(ObjectEvent):
    
    def __init__(self, object, workflow, action):
        ObjectEvent.__init__(self, object)
        self.workflow = workflow
        self.action = action
    
@implementer(IActionWillBeInvokedEvent)
class ActionWillBeInvokedEvent(WorkflowActionEvent):
    pass

            
@implementer(IActionRaisedExceptionEvent)
class ActionRaisedExceptionEvent(WorkflowActionEvent):
    pass
    
    def __init__(self, object, workflow, action, exc):
        WorkflowActionEvent.__init__(self, object, workflow, action)
        self.exc = exc
    
@implementer(IActionSucceededEvent)
class ActionSucceededEvent(WorkflowActionEvent):
    
    def __init__(self, object, workflow, action, result):
        WorkflowActionEvent.__init__(self, object, workflow, action)
        self.result = result




##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Workflow tool xml adapters and setup handlers.
"""

from zope.component import adapts
from zope.component import getSiteManager

from Products.GenericSetup.interfaces import ISetupEnviron
from Products.GenericSetup.utils import exportObjects
from Products.GenericSetup.utils import importObjects
from Products.GenericSetup.utils import ObjectManagerHelpers
from Products.GenericSetup.utils import PropertyManagerHelpers
from Products.GenericSetup.utils import XMLAdapterBase

from Products.CMFCore.interfaces import IConfigurableWorkflowTool
from Products.CMFCore.interfaces import IWorkflowTool


class WorkflowToolXMLAdapter(XMLAdapterBase, ObjectManagerHelpers,
                             PropertyManagerHelpers):

    """XML im- and exporter for WorkflowTool.
    """

    adapts(IConfigurableWorkflowTool, ISetupEnviron)

    _LOGGER_ID = 'workflow'

    name = 'workflows'

    def _exportNode(self):
        """Export the object as a DOM node.
        """
        node = self._getObjectNode('object')
        node.appendChild(self._extractProperties())
        node.appendChild(self._extractObjects())
        node.appendChild(self._extractChains())

        self._logger.info('Workflow tool exported.')
        return node

    def _importNode(self, node):
        """Import the object from the DOM node.
        """
        if self.environ.shouldPurge():
            self._purgeProperties()
            self._purgeObjects()
            self._purgeChains()

        self._initProperties(node)
        self._initObjects(node)
        self._initChains(node)

        self._logger.info('Workflow tool imported.')

    def _extractChains(self):
        fragment = self._doc.createDocumentFragment()
        node = self._doc.createElement('bindings')
        child = self._doc.createElement('default')
        chain = self.context.getDefaultChain()
        for workflow_id in chain:
            sub = self._doc.createElement('bound-workflow')
            sub.setAttribute('workflow_id', workflow_id)
            child.appendChild(sub)
        node.appendChild(child)
        for type_id, chain in self.context.listChainOverrides():
            child = self._doc.createElement('type')
            child.setAttribute('type_id', type_id)
            for workflow_id in chain:
                sub = self._doc.createElement('bound-workflow')
                sub.setAttribute('workflow_id', workflow_id)
                child.appendChild(sub)
            node.appendChild(child)
        fragment.appendChild(node)
        return fragment

    def _purgeChains(self):
        self.context.setDefaultChain('')
        for type_id, _chain in self.context.listChainOverrides():
            self.context.setChainForPortalTypes((type_id,), None,
                                                verify=False)

    def _initChains(self, node):
        for child in node.childNodes:
            if child.nodeName != 'bindings':
                continue
            for sub in child.childNodes:
                if sub.nodeName == 'default':
                    self.context.setDefaultChain(self._getChain(sub))
                if sub.nodeName == 'type':
                    type_id = str(sub.getAttribute('type_id'))
                    if sub.hasAttribute('remove'):
                        chain = None
                    else:
                        chain = self._getChain(sub)
                    self.context.setChainForPortalTypes((type_id,), chain,
                                                        verify=False)

    def _getChain(self, node):
        workflow_ids = []
        for child in node.childNodes:
            if child.nodeName != 'bound-workflow':
                continue
            workflow_ids.append(str(child.getAttribute('workflow_id')))
        return ','.join(workflow_ids)


def importWorkflowTool(context):
    """Import workflow tool and contained workflow definitions from XML files.
    """
    sm = getSiteManager(context.getSite())
    tool = sm.queryUtility(IWorkflowTool)
    if tool is None:
        logger = context.getLogger('workflow')
        logger.debug('Nothing to import.')
        return

    importObjects(tool, '', context)

def exportWorkflowTool(context):
    """Export workflow tool and contained workflow definitions as XML files.
    """
    sm = getSiteManager(context.getSite())
    tool = sm.queryUtility(IWorkflowTool)
    if tool is None:
        logger = context.getLogger('workflow')
        logger.debug('Nothing to export.')
        return

    exportObjects(tool, '', context)



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Catalog tool setup handler unit tests.
"""

import unittest
from Testing import ZopeTestCase
ZopeTestCase.installProduct('ZCTextIndex', 1)

from OFS.Folder import Folder
from Products.ZCTextIndex.Lexicon import CaseNormalizer
from Products.ZCTextIndex.Lexicon import Splitter
from Products.ZCTextIndex.Lexicon import StopWordRemover
from Products.ZCTextIndex.ZCTextIndex import PLexicon
from zope.component import getSiteManager

from Products.GenericSetup.tests.common import BaseRegistryTests
from Products.GenericSetup.tests.common import DummyExportContext
from Products.GenericSetup.tests.common import DummyImportContext

from Products.CMFCore.CatalogTool import CatalogTool
from Products.CMFCore.interfaces import ICatalogTool
from Products.CMFCore.testing import ExportImportZCMLLayer

_EMPTY_EXPORT = """\
<?xml version="1.0"?>
<object meta_type="CMF Catalog" name="portal_catalog">
 <property name="title"/>
</object>
"""

_NORMAL_EXPORT = """\
<?xml version="1.0"?>
<object meta_type="CMF Catalog" name="portal_catalog">
 <property name="title"/>
 <object name="foo_plexicon" meta_type="ZCTextIndex Lexicon">
  <element name="Whitespace splitter" group="Word Splitter"/>
  <element name="Case Normalizer" group="Case Normalizer"/>
  <element name="Remove listed stop words only" group="Stop Words"/>
 </object>
 <index name="foo_zctext" meta_type="ZCTextIndex">
  <indexed_attr value="foo_zctext"/>
  <extra name="index_type" value="Okapi BM25 Rank"/>
  <extra name="lexicon_id" value="foo_plexicon"/>
 </index>
 <column value="foo_zctext"/>
</object>
"""

_UPDATE_IMPORT = """\
<?xml version="1.0"?>
<object meta_type="CMF Catalog" name="portal_catalog">
 <index name="foo_date" meta_type="DateIndex">
  <property name="index_naive_time_as_local">True</property>
 </index>
 <column value="foo_date"/>
</object>
"""


class _extra:

    pass


class _CatalogToolSetup(BaseRegistryTests):

    def _initSite(self, foo=2):
        site = Folder(id='site').__of__(self.app)
        ctool = CatalogTool()
        getSiteManager().registerUtility(ctool, ICatalogTool)

        for obj_id in ctool.objectIds():
            ctool._delObject(obj_id)
        for idx_id in ctool.indexes():
            ctool.delIndex(idx_id)
        for col in list(ctool.schema()):
            ctool.delColumn(col)

        if foo > 0:
            ctool._setObject('foo_plexicon', PLexicon('foo_plexicon'))
            lex = ctool.foo_plexicon
            lex._pipeline = (Splitter(), CaseNormalizer(), StopWordRemover())

            extra = _extra()
            extra.lexicon_id = 'foo_plexicon'
            extra.index_type = 'Okapi BM25 Rank'
            ctool.addIndex('foo_zctext', 'ZCTextIndex', extra)

            ctool.addColumn('foo_zctext')

        return site, ctool


class exportCatalogToolTests(_CatalogToolSetup):

    layer = ExportImportZCMLLayer

    def test_unchanged(self):
        from Products.CMFCore.exportimport.catalog import exportCatalogTool

        site, _ctool = self._initSite(0)
        context = DummyExportContext(site)
        exportCatalogTool(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'catalog.xml')
        self._compareDOM(text.decode('utf8'), _EMPTY_EXPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_normal(self):
        from Products.CMFCore.exportimport.catalog import exportCatalogTool

        site, _ctool = self._initSite(2)
        context = DummyExportContext(site)
        exportCatalogTool(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'catalog.xml')
        self._compareDOM(text.decode('utf8'), _NORMAL_EXPORT)
        self.assertEqual(content_type, 'text/xml')


class importCatalogToolTests(_CatalogToolSetup):

    layer = ExportImportZCMLLayer

    def test_empty_purge(self):
        from Products.CMFCore.exportimport.catalog import importCatalogTool

        site, ctool = self._initSite(2)

        self.assertEqual(len(ctool.objectIds()), 1)
        self.assertEqual(len(ctool.indexes()), 1)
        self.assertEqual(len(ctool.schema()), 1)

        context = DummyImportContext(site, True)
        context._files['catalog.xml'] = _EMPTY_EXPORT
        importCatalogTool(context)

        self.assertEqual(len(ctool.objectIds()), 0)
        self.assertEqual(len(ctool.indexes()), 0)
        self.assertEqual(len(ctool.schema()), 0)

    def test_empty_update(self):
        from Products.CMFCore.exportimport.catalog import importCatalogTool

        site, ctool = self._initSite(2)

        self.assertEqual(len(ctool.objectIds()), 1)
        self.assertEqual(len(ctool.indexes()), 1)
        self.assertEqual(len(ctool.schema()), 1)

        context = DummyImportContext(site, False)
        context._files['catalog.xml'] = _EMPTY_EXPORT
        importCatalogTool(context)

        self.assertEqual(len(ctool.objectIds()), 1)
        self.assertEqual(len(ctool.indexes()), 1)
        self.assertEqual(len(ctool.schema()), 1)

    def test_normal_purge(self):
        from Products.CMFCore.exportimport.catalog import exportCatalogTool
        from Products.CMFCore.exportimport.catalog import importCatalogTool

        site, ctool = self._initSite(2)

        self.assertEqual(len(ctool.objectIds()), 1)
        self.assertEqual(len(ctool.indexes()), 1)
        self.assertEqual(len(ctool.schema()), 1)

        context = DummyImportContext(site, True)
        context._files['catalog.xml'] = _NORMAL_EXPORT
        importCatalogTool(context)

        self.assertEqual(len(ctool.objectIds()), 1)
        self.assertEqual(len(ctool.indexes()), 1)
        self.assertEqual(len(ctool.schema()), 1)

        # complete the roundtrip
        context = DummyExportContext(site)
        exportCatalogTool(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, 'catalog.xml')
        self._compareDOM(text.decode('utf8'), _NORMAL_EXPORT)
        self.assertEqual(content_type, 'text/xml')

    def test_normal_update(self):
        from Products.CMFCore.exportimport.catalog import importCatalogTool

        site, ctool = self._initSite(2)

        self.assertEqual(len(ctool.objectIds()), 1)
        self.assertEqual(len(ctool.indexes()), 1)
        self.assertEqual(len(ctool.schema()), 1)

        context = DummyImportContext(site, False)
        context._files['catalog.xml'] = _UPDATE_IMPORT
        importCatalogTool(context)

        self.assertEqual(len(ctool.objectIds()), 1)
        self.assertEqual(len(ctool.indexes()), 2)
        self.assertEqual(len(ctool.schema()), 2)


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(exportCatalogToolTests),
        unittest.makeSuite(importCatalogToolTests),
        ))



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for MemberDataTool module.
"""

import unittest
import Testing

import Acquisition
from AccessControl.SecurityManagement import newSecurityManager
from DateTime.DateTime import DateTime
from zope.component import getSiteManager
from zope.interface.verify import verifyClass
from zope.testing.cleanup import cleanUp

from Products.CMFCore.exceptions import BadRequest
from Products.CMFCore.interfaces import IMembershipTool
from Products.CMFCore.tests.base.security import DummyUser as BaseDummyUser


class DummyUserFolder(Acquisition.Implicit):

    def __init__(self):
        self._users = {}

    def _addUser(self, user):
        self._users[user.getId()] = user

    def userFolderEditUser(self, name, password, roles, domains):
        user = self._users[name]
        if password is not None:
            user.__ = password
        # Emulate AccessControl.User's stupid behavior (should test None)
        user._roles = tuple(roles)
        user._domains = tuple(domains)

    def getUsers(self):
        return self._users.values()


class DummyUser(BaseDummyUser):

    def __init__(self, name, password, roles, domains):
        self._id = self._name = name
        self.__ = password
        self._roles = tuple(roles)
        self._domains = tuple(domains)


class MemberDataToolTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.CMFCore.MemberDataTool import MemberDataTool

        return MemberDataTool

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def tearDown(self):
        cleanUp()

    def test_interfaces(self):
        from Products.CMFCore.interfaces import IMemberDataTool

        verifyClass(IMemberDataTool, self._getTargetClass())

    def test_new(self):
        tool = self._makeOne()
        self.assertEqual(tool.getProperty('email'), '')
        self.assertEqual(tool.getProperty('portal_skin'), '')
        self.assertEqual(tool.getProperty('listed'), False)
        self.assertEqual(tool.getProperty('login_time'),
                         DateTime('1970/01/01 00:00:00 UTC'))
        self.assertEqual(tool.getProperty('last_login_time'),
                         DateTime('1970/01/01 00:00:00 UTC'))

    def test_deleteMemberData(self):
        tool = self._makeOne()
        tool.registerMemberData('Dummy', 'user_foo')
        self.assertTrue('user_foo' in tool._members)
        self.assertTrue(tool.deleteMemberData('user_foo'))
        self.assertFalse('user_foo' in tool._members)
        self.assertFalse(tool.deleteMemberData('user_foo'))

    def test_pruneMemberData(self):
        # This needs a tad more setup
        from OFS.Folder import Folder
        from Products.CMFCore.MembershipTool import MembershipTool
        folder = Folder('test')
        folder._setObject('portal_memberdata', self._makeOne())
        sm = getSiteManager()
        sm.registerUtility(MembershipTool().__of__(folder), IMembershipTool)
        folder._setObject('acl_users', DummyUserFolder())
        tool = folder.portal_memberdata

        # Create some members
        for i in range(20):
            tool.registerMemberData('Dummy_%i' % i, 'user_foo_%i' % i)

        # None of these fake members are in the user folder, which means
        # there are 20 members and 20 "orphans"
        contents = tool.getMemberDataContents()
        info_dict = contents[0]
        self.assertEqual(info_dict['member_count'], 20)
        self.assertEqual(info_dict['orphan_count'], 20)

        # Calling the prune method should delete all orphans, so we end
        # up with no members in the tool.
        tool.pruneMemberDataContents()
        contents = tool.getMemberDataContents()
        info_dict = contents[0]
        self.assertEqual(info_dict['member_count'], 0)
        self.assertEqual(info_dict['orphan_count'], 0)


class MemberAdapterTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.CMFCore.MemberDataTool import MemberAdapter

        return MemberAdapter

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def setUp(self):
        from OFS.Folder import Folder
        from Products.CMFCore.MemberDataTool import MemberDataTool
        from Products.CMFCore.MembershipTool import MembershipTool

        self.site = Folder('test')
        self.site._setObject('portal_memberdata', MemberDataTool())
        sm = getSiteManager()
        sm.registerUtility(MembershipTool(), IMembershipTool)
        self.site._setObject('acl_users', DummyUserFolder())

    def tearDown(self):
        cleanUp()

    def test_interfaces(self):
        from AccessControl.interfaces import IUser
        from Products.CMFCore.interfaces import IMember
        from Products.CMFCore.interfaces import IMemberData

        verifyClass(IMember, self._getTargetClass())
        verifyClass(IMemberData, self._getTargetClass())
        verifyClass(IUser, self._getTargetClass())

    def test_init_does_not_persist_change(self):
        user = DummyUser('bob', 'pw', ['Role'], [])
        self._makeOne(user, self.site.portal_memberdata)
        self.assertNotIn(user.getId(), self.site.portal_memberdata._members)

    def test_notifyModified_persists_change(self):
        user = DummyUser('bob', 'pw', ['Role'], [])
        member = self._makeOne(user, self.site.portal_memberdata)
        member.notifyModified()
        self.assertIn(user.getId(), self.site.portal_memberdata._members)

    def test_setProperties(self):
        user = DummyUser('bob', 'pw', ['Role'], [])
        user = user.__of__(self.site.acl_users)
        member = self._makeOne(user, self.site.portal_memberdata)
        self.assertRaises(BadRequest, member.setProperties)

        newSecurityManager(None, DummyUser('john', 'pw', ['Role'], []))
        self.assertRaises(BadRequest, member.setProperties)

        newSecurityManager(None, user)
        member.setProperties()
        self.assertEqual(member.getProperty('email'), '')
        self.assertEqual(member.getProperty('login_time'),
                         DateTime('1970/01/01 00:00:00 UTC'))

        member.setProperties({'email': 'BOB@EXAMPLE.ORG',
                              'login_time': '2000/02/02'})
        self.assertEqual(member.getProperty('email'), 'BOB@EXAMPLE.ORG')
        self.assertEqual(member.getProperty('login_time'),
                         DateTime('2000/02/02 00:00:00'))

    def test_setSecurityProfile(self):
        user = DummyUser('bob', 'pw', ['Role'], ['domain'])
        self.site.acl_users._addUser(user)
        user = user.__of__(self.site.acl_users)
        member = self._makeOne(user, self.site.portal_memberdata)
        member.setSecurityProfile(password='newpw')
        self.assertEqual(user.__, 'newpw')
        self.assertEqual(list(user.getRoles()), ['Role'])
        self.assertEqual(list(user.getDomains()), ['domain'])
        member.setSecurityProfile(roles=['NewRole'])
        self.assertEqual(user.__, 'newpw')
        self.assertEqual(list(user.getRoles()), ['NewRole'])
        self.assertEqual(list(user.getDomains()), ['domain'])
        member.setSecurityProfile(domains=['newdomain'])
        self.assertEqual(user.__, 'newpw')
        self.assertEqual(list(user.getRoles()), ['NewRole'])
        self.assertEqual(list(user.getDomains()), ['newdomain'])


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(MemberDataToolTests),
        unittest.makeSuite(MemberAdapterTests),
        ))
